# Copyright 2023 Cristian Grosu
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import numpy as np
from sklearn.neighbors import KDTree

from .invNN import DEFAULT_MODEL_PATH, invNN
from .projections import PROJECTION_METHODS


from ..DBMInterface import DBMInterface, DBM_DEFAULT_RESOLUTION
from ..tools import get_proj_error, generate_nd_indices_parallel, euclidean

from ...utils import track_time_wrapper
from ...Logger import LoggerInterface, Logger
time_tracker_console = Logger(name="Decision Boundary Mapper - DBM", info_color="cyan", show_init=False)


class DBM(DBMInterface):
    """
    The Decision Boundary Map (DBM) class has methods for visualizing the decision boundaries of a classifier.

    Methods:
        fit: Trains the classifier on the given data set.
        get_decision_boundary_map: Returns the decision boundary map for the given classifier.
        
    Example:
        >>> from DBM import DBM
        >>> classifier = ...
        >>> dbm = DBM(classifier)
        >>> img, img_confidence, img_projection_errors, img_inverse_projection_errors, X2d_train, X2d_test = dbm.get_decision_boundary_map(X2d_train, Xnd_train, Y_train, X2d_test, Xnd_test, Y_test)
        >>> plt.imshow(img)
        >>> plt.show()
    """
    
    def __init__(self, classifier, logger:LoggerInterface=None):
        """ Initializes the DBM class.

        Args:
            classifier (tensorflow.keras.Model): The classifier to be used for the decision boundary map.
            logger (LoggerInterface, optional): The logger for outputting info messages. Defaults to console logging.
        """
        super().__init__(classifier, logger)
        self.invNN = None
        
        # creating a folder for the model
        if not os.path.exists(DEFAULT_MODEL_PATH):
           os.makedirs(DEFAULT_MODEL_PATH) 
    
    @track_time_wrapper(logger=time_tracker_console)                                             
    def fit(self, 
            X2d_train: np.ndarray, Xnd_train: np.ndarray, Y_train: np.ndarray,
            X2d_test: np.ndarray, Xnd_test: np.ndarray, Y_test: np.ndarray, 
            epochs:int=300, batch_size:int=32, 
            load_folder:str=DEFAULT_MODEL_PATH):
        """ 
        Trains the classifier on the given data set.

        Args:
            X2d_train (np.ndarray): Training data set 2D data got from the projection of the original data (e.g. PCA, t-SNE, UMAP)
            Xnd_train (np.ndarray): Training data set nD data (e.g. MNIST, CIFAR10) (i.e. the original data)
            Y_train (np.ndarray): Training data set labels
            X2d_test (np.ndarray): Testing data set 2D data got from the projection of the original data (e.g. PCA, t-SNE, UMAP)
            Xnd_test (np.ndarray): Testing data set nD data (e.g. MNIST, CIFAR10) (i.e. the original data)
            Y_test (np.ndarray): Testing data set nD data (e.g. MNIST, CIFAR10) (i.e. the original data)
            epochs (int, optional): The number of epochs for which the DBM is trained. Defaults to 300.
            batch_size (int, optional): Train batch size. Defaults to 32.
        
        Returns:
            inverse_porjection_NN (invNN): The trained inverse projection neural network.
        """
        
        inverse_projection_NN = invNN(classifier=self.classifier, folder_path=load_folder)
        inverse_projection_NN.fit(X2d_train, Xnd_train, Y_train, 
                                  X2d_test, Xnd_test, Y_test, 
                                  epochs=epochs, batch_size=batch_size)
        return inverse_projection_NN
        
    def generate_boundary_map(self, 
                              Xnd_train: np.ndarray, Y_train: np.ndarray, 
                              Xnd_test: np.ndarray, Y_test: np.ndarray,
                              X2d_train: np.ndarray = None,
                              X2d_test: np.ndarray = None,
                              train_epochs:int=300, 
                              train_batch_size:int=32,
                              resolution:int=DBM_DEFAULT_RESOLUTION,
                              use_fast_decoding:bool=False,
                              projection:str='t-SNE'                              
                              ):
        """ Generates a 2D boundary map of the classifier's decision boundary.

        Args:
            X_train (np.ndarray): Training data set
            Y_train (np.ndarray): Training data labels
            X_test (np.ndarray): Testing data set
            Y_test (np.ndarray): Testing data labels
            X2d_train (np.ndarray | None): The 2D projection of the training data. If None, the data is projected using the given projection method. Defaults to None.
            X2d_test (np.ndarray | None): The 2D projection of the testing data. If None, the data is projected using the given projection method. Defaults to None.            
            train_epochs (int, optional): The number of epochs for which the DBM is trained. Defaults to 300.
            train_batch_size (int, optional): Train batch size. Defaults to 32.
            show_predictions (bool, optional): If set to true 10 prediction examples are shown. Defaults to True.
            resolution (int, optional): _description_. Defaults to DBM_DEFAULT_RESOLUTION.
            use_fast_decoding (bool, optional): If set to true the fast decoding method is used. Defaults to False.
            projection (str, optional): The projection method to be used. Defaults to 't-SNE'.
        
        Returns:
            img (np.array): A 2D numpy array with the decision boundary map, each element is an integer representing the class of the corresponding point.
            img_confidence (np.array): A 2D numpy array with the decision boundary map, each element is a float representing the confidence of the classifier for the corresponding point.
            X2d_train (np.array): A 2D matrix representing the projection of the training data set, each element is an integer representing the class of the corresponding point.
            X2d_test (np.array): A 2D matrix representing the projection of the testing data set, each element is an integer representing the class of the corresponding point.
            space_nd (np.array): A 2D matrix representing the nD space in which the decision boundary map is generated.
            history (dict): A dictionary containing the training history of the neural network.
        
        Example:
            >>> dbm = DBM(classifier)
            >>> img, img_confidence, X2d_train, X2d_test, space_nd, history = dbm.generate_boundary_map(X_train, Y_train, X_test, Y_test)
            >>> fig, [ax1, ax2] = plt.subplots(1, 2, figsize=(10, 5))
            >>> ax1.imshow(img)
            >>> ax2.imshow(img_confidence)
            >>> plt.show()
        """
        assert projection in ['t-SNE', 'PCA', 'UMAP']
        
        if X2d_train is None or X2d_test is None:
            Xnd_train_flatten = Xnd_train.reshape((Xnd_train.shape[0], -1))
            Xnd_test_flatten = Xnd_test.reshape((Xnd_test.shape[0], -1))
            X2d_train, X2d_test = self.__transform_2d__(Xnd_train_flatten, Xnd_test_flatten, projection)
        else:
            # Normalize the data to be in the range of [0,1]
            X2d_train, X2d_test = self.__normalize_2d__(X2d_train, X2d_test)
            
        if self.invNN is None:
            self.invNN = self.fit(X2d_train, Xnd_train, Y_train, 
                                  X2d_test, Xnd_test, Y_test, 
                                  train_epochs, train_batch_size,
                                  load_folder=os.path.join(DEFAULT_MODEL_PATH, projection))   
        
        self.console.log("Decoding the 2D space... 2D -> nD")
        
        save_img_path = os.path.join(DEFAULT_MODEL_PATH, "boundary_map")
        save_img_confidence_path = os.path.join(DEFAULT_MODEL_PATH, "boundary_map_confidence")
        
        if use_fast_decoding:
            img, img_confidence, spaceNd = self._get_img_dbm_fast_(resolution)
            save_img_path += "_fast"
            save_img_confidence_path += "_fast"
        else:
            img, img_confidence, spaceNd = self._get_img_dbm_(resolution)
    
        with open(f"{save_img_path}.npy", 'wb') as f:
            np.save(f, img)
        with open(f"{save_img_confidence_path}.npy", 'wb') as f:
            np.save(f, img_confidence)        

        self.resolution = resolution   
        self.spaceNd = spaceNd     
        self.X2d = np.concatenate((X2d_train, X2d_test), axis=0)
        self.Xnd = np.concatenate((Xnd_train.reshape((Xnd_train.shape[0],-1)), Xnd_test.reshape((Xnd_test.shape[0],-1))), axis=0)
        self.console.log("Map the 2D embedding of the data to the 2D image")
        
        # transform the encoded data to be in the range [0, resolution)
        X2d_train *= (resolution - 1)
        X2d_test *= (resolution - 1)
        X2d_train = X2d_train.astype(int)
        X2d_test = X2d_test.astype(int)
        
        for [i,j] in X2d_train:
            img[i,j] = -1
            img_confidence[i,j] = 1
        for [i,j] in X2d_test:
            img[i,j] = -2
            img_confidence[i,j] = 1
            
        
        with open(os.path.join(DEFAULT_MODEL_PATH, projection, "history.json"), 'r') as f:
            history = json.load(f)
        
        return (img, img_confidence, X2d_train, X2d_test, spaceNd, history)
    
    @track_time_wrapper(logger=time_tracker_console)
    def generate_projection_errors(self, Xnd: np.ndarray = None, X2d: np.ndarray = None, resolution: int = None):
        """ Calculates the projection errors of the given data.

        Args:
            Xnd (np.array): The data to be projected. The data must be in range [0,1].
            X2d (np.array): The 2D projection of the data. The data must be in range [0,1].
            resolution (int): The resolution of the 2D space.
        Returns:
            errors (np.ndarray): The projection errors matrix of the given data. (resolution x resolution)
        """
        if resolution is None:
            if self.resolution is None:
                self.console.error("The resolution of the 2D space is not set, try to call the method 'generate_boundary_map' first.")
                raise Exception("The resolution of the 2D space is not set, try to call the method 'generate_boundary_map' first.")
            resolution = self.resolution
        if Xnd is None:
            if self.Xnd is None:
                self.console.error("The nD data is not set, try to call the method 'generate_boundary_map' first.")
                raise Exception("The nD data is not set, try to call the method 'generate_boundary_map' first.")
            Xnd = self.Xnd
        if X2d is None:
            if self.X2d is None:
                self.console.error("The 2D data is not set, try to call the method 'generate_boundary_map' first.")
                raise Exception("The 2D data is not set, try to call the method 'generate_boundary_map' first.")
            X2d = self.X2d
        
        assert len(Xnd) == len(X2d)
        
        X2d = X2d.reshape((X2d.shape[0], -1))
        Xnd = Xnd.reshape((Xnd.shape[0], -1))
        
        self.console.log("Calculating the projection errors of the given data")
        errors = np.zeros((resolution,resolution))
        
        K = 10 # Number of nearest neighbors to consider
        metric = "euclidean"
        
        self.console.log("Calculating the 2D tree")
        tree = KDTree(X2d, metric=metric)
        self.console.log("Finished computing the 2D tree")
        self.console.log("Calculating the 2D tree indices")
        indices_embedded = tree.query(X2d, k=len(X2d), return_distance=False)
        # Drop the actual point itself
        indices_embedded = indices_embedded[:, 1:]
        self.console.log("Finished computing the 2D tree indices")
        
        self.console.log("Calculating the nD distance indices")
        indices_source = generate_nd_indices_parallel(Xnd, metric=euclidean)
        self.console.log("Finished computing the nD distance indices")
        
        sparse_map = []
        for k in range(len(X2d)):
            x, y = X2d[k]
            sparse_map.append( (x, y, get_proj_error(indices_source[k], indices_embedded[k], k=K)) )            
        
        errors = self._generate_interpolation_rbf_(sparse_map, resolution, function='linear').T
        
        return errors
     
    def _predict2dspace_(self, X2d: np.ndarray):
        """ Predicts the labels for the given 2D data set.

        Args:
            X2d (np.ndarray): The 2D data set
        
        Returns:
            predicted_labels (np.ndarray): The predicted labels for the given 2D data set
            predicted_confidence (np.ndarray): The predicted probabilities for the given 2D data set
            spaceNd (np.ndarray): The decoded nD space
        """
        spaceNd, predictions = self.invNN.decode(X2d, verbose=0)
        predicted_labels = np.array([np.argmax(p) for p in predictions])
        predicted_confidence = np.array([np.max(p) for p in predictions])
        return predicted_labels, predicted_confidence, spaceNd
 
    def __transform_2d__(self, X_train: np.ndarray, X_test: np.ndarray, projection:str='t-SNE'):
        """ Transforms the given data to 2D using a projection method.

        Args:
            X_train (np.ndarray): The training data.
            X_test (np.ndarray): The test data.
            projection (str, optional): The projection method to be used. Defaults to 't-SNE'.
            
        Returns:
            X2d (np.ndarray): The transformed data in 2D.
        """
        self.console.log(f"Transforming the data to 2D using {projection}")
        
        X = np.concatenate((X_train, X_test), axis=0)
        X2d = PROJECTION_METHODS[projection](X)
            
        self.console.log(f"Finished transforming the data to 2D using {projection}")
        X2d_train = X2d[:len(X_train)]
        X2d_test = X2d[len(X_train):]
        
        # rescale to [0,1]
        X2d_train, X2d_test = self.__normalize_2d__(X2d_train, X2d_test)
        # ---------------------
        if not os.path.exists(os.path.join(DEFAULT_MODEL_PATH, projection)):
            os.makedirs(os.path.join(DEFAULT_MODEL_PATH, projection))
            
        file_path = os.path.join(DEFAULT_MODEL_PATH, projection, "train_2d.npy")
        self.console.log("Saving the 2D data to the disk: " + file_path)
        with open(file_path, 'wb') as f:
            np.save(f, X2d_train)
        
        file_path = os.path.join(DEFAULT_MODEL_PATH, projection, "test_2d.npy")
        self.console.log("Saving the 2D data to the disk: " + file_path)
        with open(file_path, 'wb') as f:
            np.save(f, X2d_test)
        
        return X2d_train, X2d_test
    
    def __normalize_2d__(self, X2d_train: np.ndarray, X2d_test: np.ndarray):
        """ Normalizes the given 2D data to [0,1].

        Args:
            X2d_train (np.ndarray): training data
            X2d_test (np.ndarray): test data

        Returns:
            X2d_train (np.ndarray): normalized training data
            X2d_test (np.ndarray): normalized test data
        """
        x_min = min(np.min(X2d_train[:,0]), np.min(X2d_test[:,0]))
        y_min = min(np.min(X2d_train[:,1]), np.min(X2d_test[:,1]))
        x_max = max(np.max(X2d_train[:,0]), np.max(X2d_test[:,0]))
        y_max = max(np.max(X2d_train[:,1]), np.max(X2d_test[:,1]))
        X2d_train = (X2d_train - np.array([x_min, y_min])) / np.array([x_max - x_min, y_max - y_min])
        X2d_test = (X2d_test - np.array([x_min, y_min])) / np.array([x_max - x_min, y_max - y_min])
        return X2d_train, X2d_test


# Copyright 2023 Cristian Grosu
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from numba import jit, njit, prange

@njit(parallel=True)
def generate_nd_indices_parallel(X_nd, metric):
    """ Generates the indices of the nearest neighbors for each point in the nD space.
        Args:
            X_nd (np.ndarray): the nD space
            metric (function): the metric to use for calculating the distances
    """
    n_samples = X_nd.shape[0]
    dist_vector = np.zeros(n_samples, dtype=np.float64)
    indices = np.zeros((n_samples, n_samples - 1), dtype=np.int64)
    
    for i in range(n_samples):
        for j in prange(n_samples):
            dist_vector[j] = metric(X_nd[i], X_nd[j])

        indices[i] = np.argsort(dist_vector)[1:] # exclude the point itself
    
    return indices

@njit(fastmath=True)
def euclidean(x, y):
    r"""Standard euclidean distance.

    ..math::
        D(x, y) = \sqrt{\sum_i (x_i - y_i)^2}
    """
    result = 0.0
    for i in range(x.shape[0]):
        result += (x[i] - y[i]) ** 2
    return np.sqrt(result)

@jit
def get_inv_proj_error(i:int,j:int, Xnd:np.ndarray, w:int=1, h:int=1):
    """Calculates the inverse projection error for a given point in the image.
        Args:
            i (int): the row of the point
            j (int): the column of the point
            Xnd (np.ndarray): the nD space
    """
    xl = Xnd[i,j-w] if j - w >= 0 else Xnd[i,j]
    xr = Xnd[i,j+w] if j + w < Xnd.shape[1] else Xnd[i,j]
    yl = Xnd[i-h,j] if i - h >= 0 else Xnd[i,j]
    yr = Xnd[i+h,j] if i + h < Xnd.shape[0] else Xnd[i,j]
    
    dw = 2 * w
    dh = 2 * h
    if (j - w < 0) or (j + w >= Xnd.shape[1]):
        dw = w
    if (i - h < 0) or (i + h >= Xnd.shape[0]):
        dh = h
    
    dx = (xl - xr) / dw
    dy = (yl - yr) / dh    
    return np.sqrt(np.linalg.norm(dx)**2 + np.linalg.norm(dy)**2)

@njit(parallel=True)
def get_proj_error(indices_source: np.ndarray, indices_embedding: np.ndarray, k: int=10):
    """ Calculates the projection error for a given data point.
        Args:
            indices_source (np.ndarray): the indices of the point neighbors in the source (i.e. nD) space
            indices_embedding (np.ndarray): the indices of the point neighbors in the embedding (i.e. 2D) space
            k (int): the number of neighbors to consider
    """    
    assert len(indices_source) == len(indices_embedding)
    n = len(indices_source)
    
    continuity = 0.0
    trustworthiness = 0.0
    
    # computing the continuity and trustworthiness errors for each point in parallel
    for i in prange(k):
        rank_2d = 0
        while indices_source[i] != indices_embedding[rank_2d]:
            rank_2d += 1

        rank_nd = 0
        while indices_source[rank_nd] != indices_embedding[i]:
            rank_nd += 1

        if rank_2d > k:
            continuity += rank_2d - k

        if rank_nd > k:
            trustworthiness += rank_nd - k


    continuity = 2 * continuity / (k * (2*n - 3*k - 1))
    trustworthiness = 2 * trustworthiness / (k * (2*n - 3*k - 1))
    
    return (continuity + trustworthiness) / 2

@jit
def get_decode_pixel_priority(img, i, j, window_size, label):
    """Calculates the priority of decoding a chunk of pixels.
       The chunk is defined by the window size and the pixel coordinates (i,j).

    Args:
        img (np.ndarray): the image w.r.t. which the priority is calculated
        i (float): the row of the current pixel
        j (float): the column of the current pixel
        window_size (float): the window size (i.e. the chunk size)
        label (any): the label of the current pixel

    Returns:
        priority (float): the priority of decoding the chunk in range [0,1]
    """
    resolution = img.shape[0]
    # getting the 4 neighbors
    i, j = int(i), int(j)
    w = int(window_size / 2)
    neighbors = []
    if i - w > 0:
        neighbors.append(img[i - w, j])
    if i + w + 1 < resolution:
        neighbors.append(img[i + w + 1, j])
    if j - w > 0 and i + 1 < resolution:
        neighbors.append(img[i + 1, j - w])
    if j + w + 1 < resolution:
        neighbors.append(img[i, j + w + 1])
        
    cost = 0
    for neighbor in neighbors:
        if neighbor != label:
            cost += 1
        
    if cost == 0:
        return -1
    
    cost /= len(neighbors)
    cost *= window_size      
    
    return 1/cost



# Copyright 2022 Cristian Grosu
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from datetime import datetime
from termcolor import colored

from .LoggerInterface import LoggerInterface

class Logger(LoggerInterface):
    """ A console logger
        Prints the messages to the default console
    """
    
    def __init__(self, active:bool=True, name:str="Logger", info_color:str="magenta", show_init:bool=True):
        """ Initialize the logger

        Args:
            active (bool, optional): Defaults to True.
            name (str, optional): Defaults to "Logger".
        """
        self.active = active
        self.name = name
        self.info_color = info_color
        
        if show_init:    
            sep = "=" * 30
            time = datetime.now().strftime("%H:%M:%S:%f")
            print(colored(f"[INFO] [{time}] [{self.name}] {sep}", self.info_color))
            print(colored(f"[INFO] [{time}] [{self.name}] {self.name} initialized", self.info_color))
            print(colored(f"[INFO] [{time}] [{self.name}] {sep}", self.info_color))
        
        
    def log(self, message:str):
        """ Log a message to the console
            Args:
                message (str): the message to log
        """
        if self.active:
            time = datetime.now().strftime("%H:%M:%S:%f")
            print(colored(f"[INFO] [{time}] [{self.name}] {message}", self.info_color))
    
    def warn(self, message:str):
        if self.active:
            time = datetime.now().strftime("%H:%M:%S:%f")
            print(colored(f"[WARNING] [{time}] [{self.name}] {message}", "yellow"))
    
    def error(self, message:str):
        if self.active:
            time = datetime.now().strftime("%H:%M:%S:%f")
            print(colored(f"[ERROR] [{time}] [{self.name}] {message}", "red"))
    
    def debug(self, message:str):
        if self.active:
            time = datetime.now().strftime("%H:%M:%S:%f")
            print(colored(f"[DEBUG] [{time}] [{self.name}] {message}","blue"))
            
    def success(self, message:str):
        if self.active:
            time = datetime.now().strftime("%H:%M:%S:%f")
            print(colored(f"[SUCCESS] [{time}] [{self.name}] {message}", "green"))


# Copyright 2023 Cristian Grosu
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import numpy as np
from sklearn.neighbors import KDTree
from math import sqrt

from .Autoencoder import DEFAULT_MODEL_PATH, Autoencoder
from ..DBMInterface import DBMInterface, DBM_DEFAULT_RESOLUTION
from ..tools import get_proj_error, generate_nd_indices_parallel, euclidean

from ...utils import track_time_wrapper
from ...Logger import LoggerInterface, Logger
time_tracker_console = Logger(name="Decision Boundary Mapper - DBM", info_color="cyan", show_init=False)


class SDBM(DBMInterface):
    """
        Self Decision Boundary Mapper (SDBM)
        Uses an auto encoder: 
        1) The encoder part of which is used to reduce the dimensionality of the data (nD -> 2D)
        2) The decoder part of which is used to reconstruct the data (2D -> nD)
        3) The classifier is used to predict the class of the data (nD -> 1D) 
        
    """
    
    def __init__(self, classifier, logger:LoggerInterface=None):
        """Initialize an Self Decision Boundary Mapper

        Args:
            classifier (tensorflow.keras.Model): The classifier to be used for the autoencoder
            logger (LoggerInterface, optional): The logger class to be used for outputting the info messages. Defaults to None.
        """
        super().__init__(classifier, logger)
        self.autoencoder = None

    @track_time_wrapper(logger=time_tracker_console)
    def fit(self, 
            X_train: np.ndarray, Y_train: np.ndarray, 
            X_test: np.ndarray, Y_test: np.ndarray, 
            epochs:int=300, batch_size:int=32,
            load_folder:str = DEFAULT_MODEL_PATH):
        """Train an autoencoder on the training data (this will be used to reduce the dimensionality of the data (nD -> 2D) and decode the 2D space to nD)

        Args:
            X_train (np.ndarray): Training data
            Y_train (np.ndarray): Training labels
            X_test (np.ndarray): Testing data
            Y_test (np.ndarray): Testing labels
            epochs (int, optional): The number of epochs to train the autoencoder. Defaults to 300.
            batch_size (int, optional): Defaults to 32.
            load_folder (str, optional): The folder path which contains a pre-trained autoencoder. Defaults to DEFAULT_MODEL_PATH.

        Returns:
            autoencoder (Autoencoder): The trained autoencoder
        """
        autoencoder = Autoencoder(folder_path = load_folder, classifier=self.classifier)
        autoencoder.fit(X_train, Y_train, X_test, Y_test, epochs, batch_size)        
        return autoencoder
    
    def generate_boundary_map(self, 
                              X_train:np.ndarray, Y_train:np.ndarray,
                              X_test:np.ndarray, Y_test:np.ndarray,
                              train_epochs:int=300, train_batch_size:int=32,
                              resolution:int=DBM_DEFAULT_RESOLUTION,
                              use_fast_decoding:bool=False,
                              projection:str = None # this parameter is not used in SDBM but is placed here to keep the same interface as DBM
                              ):
        """Generate the decision boundary map
        
            Args:
                X_train (np.ndarray): Training data
                Y_train (np.ndarray): Training labels
                X_test (np.ndarray): Testing data
                Y_test (np.ndarray): Testing labels
                train_epochs (int, optional): The number of epochs to train the autoencoder. Defaults to 300.
                train_batch_size (int, optional): Defaults to 32.
                resolution (int, optional): The resolution of the decision boundary map. Defaults to DBM_DEFAULT_RESOLUTION = 256.
                use_fast_decoding (bool, optional): If True, a fast inference algorithm will be used to decode the 2D space and to generate the decision boundary map. Defaults to False.
                projection (str, optional): The projection is not used in SDBM, is placed here just to match the DBM signature. Defaults to None.
                
            Returns:
                img (np.ndarray): The decision boundary map
                img_confidence (np.ndarray): The confidence map
                encoded_training_data (np.ndarray): The 2D coordinates of the training data for each pixel of the decision boundary map
                encdoded_testing_data (np.ndarray): The 2D coordinates of the testing data for each pixel of the decision boundary map
                space_Nd (np.ndarray): The Nd coordinates of the decision boundary map
                history (dict): The history of the training process

            Example:
                >>> import SDBM
                >>> classifier = build_classifier(...)
                >>> sdbm = SDBM.SDBM(classifier)
                >>> img, img_confidence, _, _, space_Nd, history = sdbm.generate_boundary_map(X_train, Y_train, X_test, Y_test)
                >>> plt.imshow(img)
                >>> plt.show()
        """
        
        # first train the autoencoder if it is not already trained
        if self.autoencoder is None:
            self.autoencoder = self.fit(X_train, 
                                        Y_train, 
                                        X_test, 
                                        Y_test, 
                                        epochs=train_epochs, 
                                        batch_size=train_batch_size)

        # encoder the train and test data and show the encoded data in 2D space
        self.console.log("Encoding the training data to 2D space")
        encoded_training_data = self.autoencoder.encode(X_train)
        self.console.log("Encoding the testing data to 2D space")
        encoded_testing_data = self.autoencoder.encode(X_test)            
        # generate the 2D image in the encoded space
        self.console.log("Decoding the 2D space... 2D -> nD")
        
        save_img_path = os.path.join(DEFAULT_MODEL_PATH, "boundary_map")
        save_img_confidence_path = os.path.join(DEFAULT_MODEL_PATH, "boundary_map_confidence")
        
        if use_fast_decoding:
            img, img_confidence, spaceNd = self._get_img_dbm_fast_(resolution)
            save_img_path += "_fast"
            save_img_confidence_path += "_fast"
        else:
            img, img_confidence, spaceNd = self._get_img_dbm_(resolution)
        
        self.resolution = resolution
        self.spaceNd = spaceNd
        self.X2d = np.concatenate((encoded_training_data, encoded_testing_data), axis=0)
        self.Xnd = np.concatenate((X_train.reshape((X_train.shape[0],-1)), X_test.reshape((X_test.shape[0],-1))), axis=0)
        
        # transform the encoded data to be in the range [0, resolution)
        encoded_testing_data *= (resolution -1)
        encoded_training_data *= (resolution -1)
        encoded_training_data = encoded_training_data.astype(int)
        encoded_testing_data = encoded_testing_data.astype(int)
        
        
        with open(f"{save_img_path}.npy", 'wb') as f:
            np.save(f, img)
        with open(f"{save_img_confidence_path}.npy", 'wb') as f:
            np.save(f, img_confidence)
        
        for [i,j] in encoded_training_data:
            img[i,j] = -1
            img_confidence[i,j] = 1
        for [i,j] in encoded_testing_data:
            img[i,j] = -2
            img_confidence[i,j] = 1
        
        with open(os.path.join(DEFAULT_MODEL_PATH, "history.json"), 'r') as f:
            history = json.load(f)
        
        return (img, img_confidence, encoded_training_data, encoded_testing_data, spaceNd, history)
    
    def _predict2dspace_(self, X2d:np.ndarray):
        """ Predicts the labels for the given 2D data set.

        Args:
            X2d (np.ndarray): The 2D data set
        
        Returns:
            predicted_labels (np.array): The predicted labels for the given 2D data set
            predicted_confidence (np.array): The predicted probabilities for the given 2D data set
            spaceNd (np.array): The decoded nD space
        """
        spaceNd, predictions = self.autoencoder.decode(X2d, verbose=0)
        predicted_labels = np.array([np.argmax(p) for p in predictions])
        predicted_confidence = np.array([np.max(p) for p in predictions])
        return predicted_labels, predicted_confidence, spaceNd
    
    @track_time_wrapper(logger=time_tracker_console)
    def generate_projection_errors(self, Xnd: np.ndarray = None, X2d: np.ndarray = None, resolution: int = None):
        """ Calculates the projection errors of the given data.

        Args:
            Xnd (np.array): The data to be projected. The data must be in the range [0,1].
            X2d (np.array): The 2D projection of the data. The data must be in the range [0,1].
            resolution (int): The resolution of the 2D space.
        Returns:
            errors (np.array): The projection errors matrix of the given data. 
        
        Example:
            >>> from SDBM import SDBM
            >>> classifier = ...
            >>> Xnd = ...
            >>> sdbm = SDBM(classifier)
            >>> errors = sdbm.get_projection_errors(Xnd)
            >>> plt.imshow(errors)
            >>> plt.show()
        """
        if resolution is None:
            if self.resolution is None:
                self.console.error("The resolution of the 2D space is not set, try to call the method 'generate_boundary_map' first.")
                raise Exception("The resolution of the 2D space is not set, try to call the method 'generate_boundary_map' first.")
            resolution = self.resolution
        if X2d is None:
            if self.X2d is None:
                self.console.error("No 2D data provided and no data stored in the SDBM object.")
                raise ValueError("No 2D data provided and no data stored in the SDBM object.")
            X2d = self.X2d
        if Xnd is None:
            if self.Xnd is None:
                self.console.error("No nD data provided and no data stored in the SDBM object.")
                raise ValueError("No nD data provided and no data stored in the SDBM object.")
            Xnd = self.Xnd
            
        X2d = X2d.reshape((X2d.shape[0], -1))
        Xnd = Xnd.reshape((Xnd.shape[0], -1))
        
        assert len(X2d) == len(Xnd)
        assert X2d.shape[1] == 2
        #resolution = int(sqrt(len(X2d)))
        #assert resolution * resolution == len(X2d)
        
        self.console.log("Calculating the projection errors of the given data")
        errors = np.zeros((resolution,resolution))
        
        K = 10 # Number of nearest neighbors to consider
        metric = "euclidean"
        
        self.console.log("Computing the 2D tree")
        tree = KDTree(X2d, metric=metric)
        self.console.log("Finished computing the 2D tree")
        self.console.log("Computing the 2D tree indices")
        indices_embedded = tree.query(X2d, k=len(X2d), return_distance=False)
        # Drop the actual point itself
        indices_embedded = indices_embedded[:, 1:]
        self.console.log("Finished computing the 2D tree indices")
        
        self.console.log("Calculating the nD distance indices")
        indices_source = generate_nd_indices_parallel(Xnd, metric=euclidean)
        self.console.log("Finished computing the nD distance indices")
        
        sparse_map = []
        for k in range(len(X2d)):
            x, y = X2d[k]
            sparse_map.append( (x, y, get_proj_error(indices_source[k], indices_embedded[k], k=K)) )
            
        errors = self._generate_interpolation_rbf_(sparse_map, resolution, function='linear').T
        
        return errors

    


# Copyright 2023 Cristian Grosu
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class LoggerInterface:
    """ Interface for the logger class
    """
    
    def log(self, message:str):
        """ Log a message to the console
        Args:
            message (str): the message to log
        """
        pass
        
    def warn(self, message:str):
        pass
    
    def error(self, message:str):
        pass
    
    def debug(self, message:str):
        pass
        
    def success(self, message:str):
        pass


# Copyright 2023 Cristian Grosu
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from math import sqrt
import tensorflow as tf
import os
import numpy as np
import json
import matplotlib.pyplot as plt

from ..Logger import Logger, LoggerInterface


class NNinterface:
    """ 
    Defines the interface for the neural networks used by the Decision Boundary Mapper.
    
    Methods to be implemented by the classes that inherit from this class.
    fit: Fits the model to the specified data.
    encode: Encodes the input data. (optional)
    """
    
    def __init__(self, 
                 folder_path: str,
                 classifier = None,
                 nn_name: str = "invNN", 
                 logger: LoggerInterface = None):
        """ Initializes the NN interface.

        Args:
            folder_path (str): the folder path where the model will be saved/loaded.
            classifier (tf.keras.model): The classifier model. Defaults to None.
            nn_name (str, optional): Name of the neural network that uses the classifier. Defaults to "invNN".
            logger (LoggerInterface, optional): Defaults to console logging.

        Raises:
            Exception: If the classifier is not provided.
        """
        if logger is not None:
            self.console = logger
        else:
            self.console = Logger(name=nn_name)
        
        if classifier is None:
            self.console.error("Please provide a classifier.")
            raise Exception("Please provide a classifier.")
        
        self.save_folder_path = folder_path
        self.classifier = classifier
        self.nn_name = nn_name
        self.neural_network = None
        
        try:
            self.load()
        except Exception as e:
            self.console.error("Error loading the model: {}".format(e))
            self.console.warn("The model will be built and trained from scratch.")
            
    def fit(self):
        """ Fits the model to the specified data."""
        raise NotImplementedError("The fit method is not implemented.")
    
    
    def load(self):
        """
            Loads an auto encoder from the specified folder path. With the .h5 extension.
            Args:
                folder_path (str): The path to the folder where the model is saved.
        """
        try:
            self.neural_network = tf.keras.models.load_model(os.path.join(self.save_folder_path, self.nn_name), compile=False)
            self.console.log("NN loaded successfully")
        except Exception as e:
            self.console.log(f"NN not found. Please check the path folder {self.save_folder_path} and make sure the model is saved there")        
            raise e
    
    def save(self, history):
        """
            Saves the model to the specified folder path. With the .tf extension.
        """       
        folder_path = self.save_folder_path 
        if not os.path.exists(folder_path):
            os.makedirs(folder_path)
            
        self.neural_network.save(os.path.join(folder_path, self.neural_network.name), save_format="tf")
        self.console.log(f"Model saved to {folder_path}")
        
        with open(os.path.join(folder_path, "history.json"), "w") as f:
            f.write(json.dumps(history.history))
            self.console.log(f"History saved to {folder_path}")
        
        
    def show_predictions(self, dataNd: np.ndarray, labels: np.ndarray, data2d:np.ndarray = None, n_samples:int = 16):
        """ Shows the predictions of the model. By taking first n_samples data points and comparing them to the actual labels.

        Args:
            dataNd (np.ndarray): N-dimensional data points.
            data2d (np.ndarray, optional): 2-dimensional data points. Defaults to None.
            labels (np.ndarray): Actual labels of the data points
            n_samples (int, optional): Number of samples to show. Defaults to 16.
        """
        if data2d is None:
            data2d = self.encode(dataNd)
        
        decoded, predicted = self.decode(data2d)
        predicted_labels = np.argmax(predicted, axis=1)
        
        plt.figure(figsize=(20, 10))
        m = int(sqrt(n_samples))
        n = m * m
        
        for i in range(1,2*n,2):
            plt.subplot(m, 2*m, i)
            plt.imshow(dataNd[i], cmap='gray')
            plt.title(f"Actual: {labels[i]}", color='green' if predicted_labels[i] == labels[i] else 'red', fontsize=12)
            plt.axis('off')
            plt.subplot(m, 2*m, i+1)
            plt.imshow(decoded[i], cmap='gray')
            plt.title(f"Predicted: {predicted_labels[i]}", color='green' if predicted_labels[i] == labels[i] else 'red', fontsize=12)
            plt.axis('off')
        
        plt.show()
    
    def decode(self, data: np.ndarray, verbose: int = 0):
        """ Decodes the data points.

        Args:
            data (np.ndarray): The data points to decode.
            verbose (int, optional): Verbosity mode. Defaults to 0.
            
        Returns:
            Xnd, labels_predictions: The decoded data points and the predictions of the classifier.
        """
        return self.neural_network.predict(data, verbose=verbose)
    
    def encode(self, data: np.ndarray, verbose: int = 0):
        """ Encodes the data points.

        Args:
            data (np.ndarray): The data points to encode.
            verbose (int, optional): Verbosity mode. Defaults to 0.

        Returns:
            np.ndarray: The encoded 2D data points.
        """
        return NotImplementedError("The encode method is not implemented.")


# Copyright 2023 Cristian Grosu
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from termcolor import colored

from .LoggerInterface import LoggerInterface

class LoggerGUI(LoggerInterface):
    """ Logs the messages to the GUI.
    """
    
    def __init__(self, name:str="Logger", active:bool=True, output=None, update_callback=lambda x: x, info_color:str="magenta", show_init:bool=True):
        """ Initialize the logger

        Args:
            name (str, optional): Defaults to "Logger".
            active (bool, optional): Defaults to True.
            output (any, optional): The GUI reference where the logger should display messages. Defaults to None.
            update_callback (python function, optional): The function that is called to update the GUI. Defaults to lambda x:x.
        """
        super().__init__()

        self.active = active 
        self.name = name
        self.info_color = info_color
    
        if output is not None:
            self.output = output
            self.update_callback = update_callback
        else:
            print("No output provided for LoggerGUI. LoggerGUI will not print anything.")
            self.active = False
        
        if show_init and self.active:    
            sep = "=" * 30
            time = datetime.now().strftime("%H:%M:%S:%f")
            
            self.print(f"[INFO] [{time}] [{self.name}] {sep}", self.info_color)
            self.print(f"[INFO] [{time}] [{self.name}] {self.name} initialized", self.info_color)
            self.print(f"[INFO] [{time}] [{self.name}] {sep}", self.info_color)


    def print(self, message:str, color:str='magenta'):
        if self.active:
            self.output.print(message, text_color=color)
            self.update_callback()
            
    def log(self, message:str):
        time = datetime.now().strftime("%H:%M:%S:%f")
        self.print(f"[INFO] [{time}] [{self.name}] {message}", self.info_color)
    
    def warn(self, message:str):
        time = datetime.now().strftime("%H:%M:%S:%f")
        self.print(f"[WARNING] [{time}] [{self.name}] {message}", "yellow")
    
    
    def error(self, message:str):
        time = datetime.now().strftime("%H:%M:%S:%f")
        self.print(f"[ERROR] [{time}] [{self.name}] {message}", "red")        
    
    def debug(self, message:str):
        time = datetime.now().strftime("%H:%M:%S:%f")
        self.print(f"[DEBUG] [{time}] [{self.name}] {message}","blue")
    
        
    def success(self, message:str):
        time = datetime.now().strftime("%H:%M:%S:%f")
        self.print(f"[SUCCESS] [{time}] [{self.name}] {message}", "green")


# Copyright 2023 Cristian Grosu
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tensorflow as tf
import numpy as np

from ..NNinterface import NNinterface
from ...Logger import LoggerInterface, LoggerModel

DEFAULT_MODEL_PATH = os.path.join("tmp", "SDBM")
DECODER_NAME = "decoder"
ENCODER_NAME = "encoder"
AUTOENCODER_NAME = "autoencoder"

class Autoencoder(NNinterface):
    def __init__(self, 
                 classifier = None, 
                 folder_path:str = DEFAULT_MODEL_PATH, 
                 logger:LoggerInterface = None):
        """
            Creates an autoencoder model.
            Classifier: The classifier part of the autoencoder.
            
            Args:
                folder_path: The path to the folder where the model will be saved/loaded.
                logger: The logger used to log the model's progress.
                classifier: The classifier part of the autoencoder.
        """
        super().__init__(folder_path=folder_path, logger=logger, nn_name=AUTOENCODER_NAME, classifier=classifier)
    
    
    def __build__(self, input_shape:tuple = (28, 28), show_summary:bool = False):
        """
            Assembles the autoencoder model and compiles it. 
            
            Args:
                input_shape: The input and output shape of the autoencoder.
                show_summary: If True, the model summary will be printed.
        """
        
        encoder = tf.keras.models.Sequential([
            tf.keras.layers.Flatten(),
            tf.keras.layers.Dense(512, activation='relu', kernel_initializer='he_uniform', kernel_regularizer=tf.keras.regularizers.l2(0.0002)),
            tf.keras.layers.Dense(128, activation='relu', kernel_initializer='he_uniform', bias_initializer=tf.keras.initializers.Constant(0.01)),
            tf.keras.layers.Dense(64, activation='relu', kernel_initializer='he_uniform', bias_initializer=tf.keras.initializers.Constant(0.01)),
            tf.keras.layers.Dense(32, activation='relu', kernel_initializer='he_uniform', bias_initializer=tf.keras.initializers.Constant(0.01)),
            tf.keras.layers.Dense(2, activation='sigmoid', bias_initializer=tf.keras.initializers.Constant(0.01)),            
        ], name=ENCODER_NAME)

        decoder = tf.keras.models.Sequential([
            tf.keras.layers.Dense(32, activation='relu', kernel_initializer='he_uniform', kernel_regularizer=tf.keras.regularizers.l2(0.0002)),
            tf.keras.layers.Dense(64, activation='relu', kernel_initializer='he_uniform', bias_initializer=tf.keras.initializers.Constant(0.01)),
            tf.keras.layers.Dense(128, activation='relu', kernel_initializer='he_uniform', bias_initializer=tf.keras.initializers.Constant(0.01)),
            tf.keras.layers.Dense(512, activation='relu', kernel_initializer='he_uniform', bias_initializer=tf.keras.initializers.Constant(0.01)),
            tf.keras.layers.Dense(input_shape[0] * input_shape[1], activation='sigmoid'),
            tf.keras.layers.Reshape(input_shape)
        ], name=DECODER_NAME)       
        
        
        input_layer = tf.keras.Input(shape=input_shape, name="input")
        
        CLASSIFIER_NAME = self.classifier.name
        DECODER_LOSS = "mean_squared_error"
        CLASSIFIER_LOSS = "sparse_categorical_crossentropy"
        DECODER_LOSS_WEIGHT = 1.0
        CLASSIFIER_LOSS_WEIGHT = 0.125
        
        encoder_output = encoder(input_layer)
        decoder_output = decoder(encoder_output)
        classifier_output = self.classifier(decoder_output)
        
        self.neural_network = tf.keras.models.Model(inputs=input_layer,
                                                    outputs=[decoder_output, classifier_output],
                                                    name=AUTOENCODER_NAME)
        
        self.neural_network.compile(optimizer=tf.keras.optimizers.Adam(), 
                                    loss = {DECODER_NAME:DECODER_LOSS,
                                            CLASSIFIER_NAME:CLASSIFIER_LOSS},
                                    loss_weights = {DECODER_NAME:DECODER_LOSS_WEIGHT,
                                                    CLASSIFIER_NAME:CLASSIFIER_LOSS_WEIGHT},
                                    metrics = {DECODER_NAME: "accuracy", CLASSIFIER_NAME: "accuracy"})
        
        if show_summary:
            self.neural_network.summary()
        
        
    def fit(self, x_train: np.ndarray, y_train: np.ndarray, 
            x_test: np.ndarray, y_test: np.ndarray,
            epochs:int = 10, batch_size:int = 128):
        """ Fits the model to the specified data.

        Args:
            x_train (np.ndarray): Train input values
            y_train (np.ndarray): Train target values
            x_test (np.ndarray): Test input values
            y_test (np.ndarray): Test target values
            epochs (int, optional): The number of epochs. Defaults to 10.
            batch_size (int, optional): Data points used for one batch. Defaults to 128.
        """
        if self.neural_network is not None:
            self.console.log("Model already loaded. Skipping build.")
            self.encoder = self.neural_network.get_layer(ENCODER_NAME)
            self.decoder = self.neural_network.get_layer(DECODER_NAME)
            self.classifier = self.neural_network.get_layer(self.classifier.name)
            return
        
        self.console.log("Building model according to the data shape.")
        self.__build__(input_shape=x_train.shape[1:], show_summary=True)
            
        stopping_callback = tf.keras.callbacks.EarlyStopping(monitor='val_loss', min_delta=0.00001, mode='min', patience=20, restore_best_weights=True)
        logger_callback = LoggerModel(name=AUTOENCODER_NAME, show_init=False, epochs=epochs)

        self.console.log("Fitting model...")
        
        history = self.neural_network.fit(x_train, [x_train, y_train], 
                            epochs=epochs, 
                            batch_size=batch_size, 
                            shuffle=True,
                            validation_data=(x_test, [x_test, y_test]),
                            callbacks=[stopping_callback, logger_callback],
                            verbose=0)
        
        self.console.log("Model fitted!")
        self.save(history)    
        self.encoder = self.neural_network.get_layer(ENCODER_NAME)
        self.decoder = self.neural_network.get_layer(DECODER_NAME)
        self.classifier = self.neural_network.get_layer(self.classifier.name)
            
        #self.show_predictions(dataNd=x_test, labels=y_test)
            
    def encode(self, data:np.ndarray, verbose:int = 0):
        """ Encodes the data using the encoder part of the autoencoder.

        Args:
            data (np.ndarray): The data to encode.
            verbose (int, optional): Verbosity level. Defaults to 0.

        Returns:
            np.ndarray: The encoded 2D data.
        """
        return self.encoder.predict(data, verbose=verbose)
        
    def decode(self, data:np.ndarray, verbose:int = 0):
        """ Decodes the data using the decoder part of the autoencoder.

        Args:
            data (np.ndarray): The data to decode.
            verbose (int, optional): Verbosity level. Defaults to 0.

        Returns:
            np.ndarray: The decoded nD data and classifier predictions.
        """
        xNd = self.decoder.predict(data, verbose=verbose)
        predictions = self.classifier.predict(xNd, verbose=verbose)
        return xNd, predictions
    


# Copyright 2023 Cristian Grosu
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf
import numpy as np
import os

from ..NNinterface import NNinterface
from ...Logger import LoggerInterface, LoggerModel

DEFAULT_MODEL_PATH = os.path.join("tmp", "DBM")
INVNN_NAME = "invNN"
DECODER_NAME = "decoder"

class invNN(NNinterface):
    """
        Inverse Projection Neural Network.
        The network is composed of 2 parts:
            1) The decoder part, i.e. The inverse projection 2D -> nD. (Sequential model 2 -> 32 -> 64 -> 128 -> 512 -> nD)
            2) The classifier part nD -> 1D (provided by the user).
    """
    
    def __init__(self, 
                 classifier = None, 
                 logger: LoggerInterface = None, 
                 folder_path: str = DEFAULT_MODEL_PATH):
        """
            Creates an inverse Projection Neural Network model.
            Classifier: The classifier part nD -> 1D.
        """
        super().__init__(folder_path = folder_path, classifier = classifier, nn_name=INVNN_NAME, logger=logger)
        
    
    def __build__(self, output_shape:tuple = (2,2), show_summary:bool = False):
        """Builds an invNN (Sequential 2D -> 32 -> 64 -> 128 -> 512 -> nD -> ... -> 1D)

        Args:
            output_shape (tuple, optional): The output shape of the Nd data. Defaults to (2,2).
            show_summary (bool, optional): If True, the model summary will be printed. Defaults to False.
        """
        
        assert len(output_shape) == 2, "Output shape must be a 2D tuple"
        
        CLASSIFIER_NAME = self.classifier.name
        DECODER_LOSS = "mean_squared_error"
        CLASSIFIER_LOSS = "sparse_categorical_crossentropy"
        DECODER_LOSS_WEIGHT = 1.0
        CLASSIFIER_LOSS_WEIGHT = 0.125
                  
        self.decoder = tf.keras.Sequential([
            tf.keras.layers.Dense(32, activation='relu', kernel_initializer='he_uniform', kernel_regularizer=tf.keras.regularizers.l2(0.0002)),
            tf.keras.layers.Dense(64, activation='relu', kernel_initializer='he_uniform', bias_initializer=tf.keras.initializers.Constant(0.01)),
            tf.keras.layers.Dense(128, activation='relu', kernel_initializer='he_uniform', bias_initializer=tf.keras.initializers.Constant(0.01)),
            tf.keras.layers.Dense(512, activation='relu', kernel_initializer='he_uniform', bias_initializer=tf.keras.initializers.Constant(0.01)),
            tf.keras.layers.Dense(output_shape[0] * output_shape[1], activation='sigmoid', kernel_initializer='he_uniform'),
            tf.keras.layers.Reshape(output_shape)
        ], name=DECODER_NAME)
        
        input_layer = tf.keras.Input(shape=(2,), name="input")
                                             
        decoder_output = self.decoder(input_layer)
        decoder_classifier_output = self.classifier(decoder_output)
        
            
        self.neural_network = tf.keras.models.Model(inputs=input_layer, 
                                            outputs=[decoder_output, 
                                                    decoder_classifier_output], 
                                            name=INVNN_NAME)
        
        self.neural_network.compile(optimizer=tf.keras.optimizers.Adam(), 
                            loss={DECODER_NAME: DECODER_LOSS,
                                  CLASSIFIER_NAME: CLASSIFIER_LOSS},
                            loss_weights={DECODER_NAME: DECODER_LOSS_WEIGHT,
                                          CLASSIFIER_NAME: CLASSIFIER_LOSS_WEIGHT},
                            metrics={DECODER_NAME: "accuracy", 
                                     CLASSIFIER_NAME: "accuracy"})
        if show_summary:
            self.neural_network.summary()
        
    def fit(self, 
            x2d_train: np.ndarray, xNd_train: np.ndarray, y_train: np.ndarray, 
            x2d_test: np.ndarray, xNd_test: np.ndarray, y_test: np.ndarray, 
            epochs:int = 300, batch_size:int = 32):
        """ Fits the model to the specified data.

        Args:
            x2d_train (np.ndarray): Train input values (2D)
            xNd_train (np.ndarray): Train input values (nD)
            y_train (np.ndarray): Train target values
            x2d_test (np.ndarray): Test input values (2D)
            xNd_test (np.ndarray): Test input values (nD)
            y_test (np.ndarray): Test target values
            epochs (int, optional): The number of epochs. Defaults to 300.
            batch_size (int, optional): Data points used for one batch. Defaults to 32.
        """
        if self.neural_network is not None:
            self.console.log("Model already loaded. Skipping build.")
            return
        
        self.console.log("Building model according to the data shape.")
        self.__build__(output_shape=xNd_train.shape[1:], show_summary=True)
            
        stopping_callback = tf.keras.callbacks.EarlyStopping(monitor='val_loss', min_delta=0.00001, mode='min', patience=20, restore_best_weights=True)
        logger_callback = LoggerModel(name=INVNN_NAME, show_init=False, epochs=epochs)
        self.console.log("Fitting model...")

        hist = self.neural_network.fit(x2d_train, [xNd_train, y_train], 
                                epochs=epochs, 
                                batch_size=batch_size, 
                                shuffle=True,
                                validation_data=(x2d_test, [xNd_test, y_test]),
                                callbacks=[stopping_callback, logger_callback],
                                verbose=0)

        self.console.log("Model fitted!")
        self.save(hist)    
        #self.show_predictions(dataNd=xNd_test, data2d=x2d_test, labels=y_test)


# Copyright 2023 Cristian Grosu
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from shutil import rmtree
import matplotlib.pyplot as plt
import PySimpleGUI as sg
import numpy as np
from PIL import Image, ImageTk
import tensorflow as tf
from tensorflow.keras.utils import plot_model

from .DBMPlotterGUI import DBMPlotterGUI
from ..Logger import LoggerGUI, Logger
from ..DBM import SDBM, DBM
from ..utils import import_csv_dataset, import_mnist_dataset

sg.theme('DarkBlue1')
TITLE = "Classifiers visualization tool"
WINDOW_SIZE = (1150, 650)
BLACK_COLOR = "#252526"
BUTTON_PRIMARY_COLOR = "#007acc"
WHITE_COLOR = "#ffffff"
RIGHTS_MESSAGE = "© 2023 Cristian Grosu. All rights reserved."
RIGHTS_MESSAGE_2 = "Made by Cristian Grosu for Utrecht University Master Thesis in 2023"

DEFAULT_DBM_IMAGE_PATH = os.path.join(os.getcwd(), "results", "MNIST", "2D_boundary_mapping.png") # unused
TMP_FOLDER = os.path.join(os.getcwd(), "tmp")

SAMPLES_LIMIT = 5000 # Limit the number of samples to be loaded from the dataset

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2' # Disable tensorflow logs

"""
0 = all messages are logged (default behavior)
1 = INFO messages are not printed
2 = INFO and WARNING messages are not printed
3 = INFO, WARNING, and ERROR messages are not printed
"""

DBM_TECHNIQUES = {
    "Autoencoder": SDBM,
    "Inverse Projection": DBM,
}

PROJECTION_TECHNIQUES = [
    "t-SNE",
    "UMAP",
    "PCA",
]

class GUI:
    def __init__(self):
        self.create_tmp_folder()
        self.window = self.build()
        self.logger = Logger(name = "GUI")
            
        # --------------- DBM ---------------
        self.dbm_plotter_gui = None
        self.dbm_logger = LoggerGUI(name = "DBM logger", output = self.window["-LOGGER-"], update_callback = self.window.refresh)
    
    def create_tmp_folder(self):
        if not os.path.exists(TMP_FOLDER):
            os.makedirs(TMP_FOLDER)
    
    def remove_tmp_folder(self):
        if os.path.exists(TMP_FOLDER):
            rmtree(TMP_FOLDER)
        
    def upload_classifier(self):
        self.classifier = tf.keras.models.Sequential([
            tf.keras.layers.Flatten(),
            tf.keras.layers.Dense(self.num_classes, activation=tf.nn.softmax)
        ], name="classifier")
        input_shape = self.X_train.shape[1:]
        input_shape = (None, *input_shape)
        self.classifier.build(input_shape=input_shape)
        file = os.path.join(TMP_FOLDER, "classifier.png")
        plot_model(self.classifier, to_file=file, show_shapes=True, show_layer_names=True, show_layer_activations=True)    
        img = Image.open(file)
        
        img.thumbnail((256, 256), Image.ANTIALIAS)           
        image = ImageTk.PhotoImage(img)     
        self.window["-CLASSIFIER IMAGE-"].update(data = image)
        self.switch_visibility(["-CLASSIFIER IMAGE-", "-CLASSIFIER TEXT-"], True)
        self.window.refresh()
    
    def build(self):
        window = sg.Window(TITLE, 
                           layout=self._get_layout(), 
                           size=WINDOW_SIZE,                            
                           resizable=False,
                           )
        window.finalize()
        return window
        
    def start(self):
        while True:
            event, values = self.window.read()
            
            if event == "Exit" or event == sg.WIN_CLOSED:
                break
            
            self.handle_event(event, values)
        
        self.stop()
    
    def stop(self):
        #self.logger.log("Removing tmp folder...")
        #self.remove_tmp_folder()
        self.logger.log("Closing the application...")
        self.window.close()
    
    def handle_event(self, event, values):        
        # Folder name was filled in, make a list of files in the folder
        EVENTS = {
            "-FOLDER-": self.handle_select_folder_event,
            "-FILE LIST-": self.handle_file_list_event,
            "-DBM TECHNIQUE-": self.handle_dbm_technique_event,
            "-DBM BTN-": self.handle_get_decision_boundary_mapping_event,
            "-DBM IMAGE-": self.handle_dbm_image_event,
            "-PROJECTION TECHNIQUE-": self.handle_projection_technique_event,
            "-UPLOAD TRAIN DATA BTN-": self.handle_upload_train_data_event,
            "-UPLOAD TEST DATA BTN-": self.handle_upload_test_data_event,
            "-UPLOAD MNIST DATA BTN-": self.handle_upload_mnist_data_event,
        }
        
        EVENTS[event](event, values)
        
    def _get_layout(self):                
        data_files_list_column = [
            [
                sg.Text(text = "Data Folder"),
                sg.In(enable_events=True, key="-FOLDER-", background_color=WHITE_COLOR, text_color=BLACK_COLOR, expand_x=True),
                sg.FolderBrowse(button_text="Browse folder", button_color=(WHITE_COLOR, BUTTON_PRIMARY_COLOR), initial_folder=os.getcwd()),
            ],
            [
               sg.Text("Choose the data file from the list: ", expand_x=True),
            ],
            [   
               sg.Text(key="-TOUT-", expand_x=True),            
            ],
            [
                sg.Listbox(
                    values=[], enable_events=True, key="-FILE LIST-", background_color=WHITE_COLOR, text_color=BLACK_COLOR, expand_x=True, expand_y=True
                )
            ],
            [
                sg.Button("Upload train data for DBM", button_color=(WHITE_COLOR, BUTTON_PRIMARY_COLOR), expand_x=True, key = "-UPLOAD TRAIN DATA BTN-"),
                sg.Button("Upload test data for DBM", button_color=(WHITE_COLOR, BUTTON_PRIMARY_COLOR), expand_x=True, key = "-UPLOAD TEST DATA BTN-"),
            ],
            [   
               sg.Text("Training data file: ", key="-TRAIN DATA FILE-",  expand_x=True),            
            ],
            [   
               sg.Text("Training data shape: ", key="-TRAIN DATA SHAPE-", expand_x=True),            
            ],
            [   
               sg.Text("Testing data file: ", key="-TEST DATA FILE-", expand_x=True),            
            ],
            [   
               sg.Text("Testing data shape: ", key="-TEST DATA SHAPE-", expand_x=True),            
            ],
            [
                sg.Button("Upload MNIST Data set", button_color=(WHITE_COLOR, BUTTON_PRIMARY_COLOR), expand_x=True, key = "-UPLOAD MNIST DATA BTN-"),
            ],
            [
                sg.Button("Show the Decision Boundary Mapping", button_color=(WHITE_COLOR, BUTTON_PRIMARY_COLOR), expand_x=True, key = "-DBM BTN-"),
            ],
            [
                sg.Text("", expand_x=True)    
            ],
            [
                sg.Text(RIGHTS_MESSAGE, expand_x=True),
            ],
            [
                sg.Text(RIGHTS_MESSAGE_2, expand_x=True),
            ]
        ]
        
        results_column = [
            [
                sg.Text("Which dbm technique would you like to use?", size=(45,1)),
                sg.Combo(
                    values=list(DBM_TECHNIQUES.keys()),
                    default_value=list(DBM_TECHNIQUES.keys())[0],
                    expand_x=True,
                    key="-DBM TECHNIQUE-",
                    enable_events=True,
                    background_color=WHITE_COLOR, text_color=BLACK_COLOR,
                ),
            ],
            [
                sg.Text("Which Projection technique would you like to use?", size=(45,1), key="-PROJECTION TECHNIQUE TEXT-", visible=False),
                sg.Combo(
                    values = PROJECTION_TECHNIQUES,
                    default_value = PROJECTION_TECHNIQUES[0],
                    expand_x=True,
                    key="-PROJECTION TECHNIQUE-",
                    enable_events=True,
                    visible=False,
                    background_color=WHITE_COLOR, text_color=BLACK_COLOR,
                ),
            ],
            [
                sg.Text("Use the fast algorithm for generating the Decision Boundary Mapper: ", expand_x=True, key="-USE FAST DBM TEXT-", visible=True),
                sg.Checkbox("", default=False,  key="-USE FAST DBM CHECKBOX-", visible=True),
            ],            
            [
                sg.Text("Show Decision Boundary Mapper NN history: ", expand_x=True, key="-DBM HISTORY TEXT-", visible=True),
                sg.Checkbox("", default=False,  key="-DBM HISTORY CHECKBOX-", visible=True),
            ],
            # ---------------------------------------------------------------------------------------------------                        
            [
                sg.Column([
                    [sg.Text("Classifier: ",  visible=False, expand_x=True, key="-CLASSIFIER TEXT-", justification='center')],
                    [sg.Image(key="-CLASSIFIER IMAGE-", expand_x=True, expand_y=True, visible=False, enable_events=True)],                           
                ], expand_x=True),
                sg.Column([
                    [sg.Text("Decision boundary map: ", visible=False, expand_x=True, key="-DBM TEXT-", justification='center')],
                    [sg.Text("Loading... ",  visible=False, expand_x=True, key="-DBM IMAGE LOADING-", justification='center')],
                    [sg.Image(key="-DBM IMAGE-", visible=False, expand_x=True, expand_y=True, enable_events=True)],                                           
                ], expand_x=True),
            ],            
            [
                sg.HSeparator(),
            ],
            [
                sg.Column([
                    [sg.Text("Logger: ")],
                    [sg.Multiline("",expand_x=True, expand_y=True, key="-LOGGER-", background_color=WHITE_COLOR, text_color=BLACK_COLOR, auto_size_text=True)],   
                ], expand_x=True, expand_y=True),    
            ]
        ]
        
        
        layout = [
            [
                sg.Column(data_files_list_column, expand_x=True, expand_y=True),   
                sg.VSeparator(),             
                sg.Column(results_column, expand_x=True, expand_y=True),
            ],
        ]

        return layout

    def handle_select_folder_event(self, event, values):
        folder = values["-FOLDER-"]
        try:
            # Get list of files in folder
            file_list = os.listdir(folder)
        except:
            file_list = []

        fnames = [ f for f in file_list
                   if os.path.isfile(os.path.join(folder, f)) and (f.lower().endswith((".csv")) or f.lower().endswith((".txt")))
                ]
        self.window["-FILE LIST-"].update(fnames)
        
    def handle_file_list_event(self, event, values):
            try:
                filename = os.path.join(
                    values["-FOLDER-"], values["-FILE LIST-"][0]
                )
                self.window["-TOUT-"].update(filename)
            except Exception as e:
                self.logger.error("Error while loading data file" + str(e))
    
    def handle_projection_technique_event(self, event, values):
        projection_technique = values["-PROJECTION TECHNIQUE-"]
        self.logger.log(f"Projection technique: {projection_technique}")
    
    def handle_dbm_technique_event(self, event, values):
        dbm_technique = values["-DBM TECHNIQUE-"]
        if dbm_technique == "Inverse Projection":
            self.switch_visibility(["-PROJECTION TECHNIQUE TEXT-", "-PROJECTION TECHNIQUE-"], True)
        else:
            self.switch_visibility(["-PROJECTION TECHNIQUE TEXT-", "-PROJECTION TECHNIQUE-"], False)
            
        self.logger.log(f"DBM technique: {dbm_technique}")
    
    def handle_upload_train_data_event(self, event, values):
        try:
            filename = os.path.join(
                values["-FOLDER-"], values["-FILE LIST-"][0]
            )
            self.X_train, self.Y_train = import_csv_dataset(filename, limit=int(0.7*SAMPLES_LIMIT))
            
            self.num_classes = np.unique(self.Y_train).shape[0]
            self.upload_classifier()
        
            self.window["-TRAIN DATA FILE-"].update("Training data file: " + filename)
            self.window["-TRAIN DATA SHAPE-"].update(f"Training data shape: X {self.X_train.shape} Y {self.Y_train.shape}")
        except Exception as e:
            self.logger.error("Error while loading data file" + str(e))
            
    def handle_upload_test_data_event(self, event, values):
        try:
            filename = os.path.join(
                values["-FOLDER-"], values["-FILE LIST-"][0]
            )
            self.X_test, self.Y_test = import_csv_dataset(filename, limit=int(0.3*SAMPLES_LIMIT))
            self.window["-TEST DATA FILE-"].update("Testing data file: " + filename)
            self.window["-TEST DATA SHAPE-"].update(f"Testing data shape: X {self.X_test.shape} Y {self.Y_test.shape}")
        except Exception as e:
            self.logger.error("Error while loading data file" + str(e))
    
    def handle_upload_mnist_data_event(self, event, values):
        (X_train, Y_train), (X_test, Y_test) = import_mnist_dataset()
        
        X_train = X_train.astype('float32')
        X_test = X_test.astype('float32')
        X_train /= 255
        X_test /= 255
        X_train, Y_train = X_train[:int(0.7*SAMPLES_LIMIT)], Y_train[:int(0.7*SAMPLES_LIMIT)]
        X_test, Y_test = X_test[:int(0.3*SAMPLES_LIMIT)], Y_test[:int(0.3*SAMPLES_LIMIT)]
        
        self.X_train, self.Y_train, self.X_test, self.Y_test = X_train, Y_train, X_test, Y_test
        self.num_classes = np.unique(self.Y_train).shape[0]
        
        self.window["-TRAIN DATA FILE-"].update("Training data: MNIST")
        self.window["-TRAIN DATA SHAPE-"].update(f"Training data shape: X {self.X_train.shape} Y {self.Y_train.shape}")
        
        self.window["-TEST DATA FILE-"].update("Testing data: MNIST")
        self.window["-TEST DATA SHAPE-"].update(f"Testing data shape: X {self.X_test.shape} Y {self.Y_test.shape}")
        
        self.upload_classifier()
    
    def switch_visibility(self, elements, visible):
        for x in elements:
            self.window[x].update(visible=visible)
        self.window.refresh()
        
    def handle_dbm_image_event(self, event, values):
        self.logger.log("Clicked on the dbm image")
        if self.dbm_plotter_gui is None:
            self.logger.warn("Nothing to show")
            return
        self.dbm_plotter_gui.start()
        
    def handle_show_dbm_history(self, training_history):
        # this is for plotting the training history
        fig, [ax1, ax2, ax3] = plt.subplots(1, 3, figsize=(20, 5))
        ax1.set_title("Loss")
        ax1.plot(training_history["loss"], label="loss")
        ax1.plot(training_history["val_loss"], label="val_loss")
        ax1.legend()
        
        ax2.set_title("Decoder Accuracy")
        ax2.plot(training_history["decoder_accuracy"], label="decoder_accuracy")
        ax2.plot(training_history["val_decoder_accuracy"], label="val_decoder_accuracy")
        ax2.legend()
        
        ax3.set_title("Classifier Accuracy")
        ax3.plot(training_history["classifier_accuracy"], label="classifier_accuracy")
        ax3.plot(training_history["val_classifier_accuracy"], label="val_classifier_accuracy")
        ax3.legend()
        
        plt.show()
            
    def handle_get_decision_boundary_mapping_event(self, event, values):
        # update loading state
        self.switch_visibility(["-DBM IMAGE-"], False)
        self.switch_visibility(["-DBM TEXT-", "-DBM IMAGE LOADING-"], True)
        
        dbm = DBM_TECHNIQUES[values["-DBM TECHNIQUE-"]](classifier = self.classifier, logger = self.dbm_logger)
        
        projection_technique = values["-PROJECTION TECHNIQUE-"]
        use_decoding_fast = values["-USE FAST DBM CHECKBOX-"]
        show_dbm_history = values["-DBM HISTORY CHECKBOX-"]
        
        RESOLUTION = 256
        
        if values["-DBM TECHNIQUE-"] == "Inverse Projection":
            DEFAULT_MODEL_FOLDER = os.path.join("tmp", "DBM")        
            
            if not os.path.exists(os.path.join(DEFAULT_MODEL_FOLDER, projection_technique, "train_2d.npy")):
                X_train_2d = None
            else:
                with open(os.path.join(DEFAULT_MODEL_FOLDER, projection_technique, "train_2d.npy"), "rb") as f:
                    X_train_2d = np.load(f)
            if not os.path.exists(os.path.join(DEFAULT_MODEL_FOLDER, projection_technique, "test_2d.npy")):
                X_test_2d = None
            else:
                with open(os.path.join(DEFAULT_MODEL_FOLDER, projection_technique, "test_2d.npy"), "rb") as f:
                    X_test_2d = np.load(f)
               
            dbm_info = dbm.generate_boundary_map(
                                        self.X_train, 
                                        self.Y_train, 
                                        self.X_test, 
                                        self.Y_test, 
                                        X2d_train=X_train_2d,
                                        X2d_test=X_test_2d,
                                        resolution=RESOLUTION,
                                        use_fast_decoding=use_decoding_fast,
                                        projection=projection_technique
                                        )
        else:
            dbm_info = dbm.generate_boundary_map(
                                        self.X_train, 
                                        self.Y_train, 
                                        self.X_test, 
                                        self.Y_test, 
                                        resolution=RESOLUTION,
                                        use_fast_decoding=use_decoding_fast,
                                        projection=projection_technique
                                        )

        img, img_confidence, encoded_training_data, encoded_testing_data, spaceNd, training_history = dbm_info
        
        if show_dbm_history:
            self.handle_show_dbm_history(training_history)
        
        
        # ---------------------------------
        # create a GUI window for Decision Boundary Mapping
        self.dbm_plotter_gui = DBMPlotterGUI(
                                            dbm_model = dbm,
                                            img = img,
                                            img_confidence = img_confidence,
                                            encoded_train = encoded_training_data, 
                                            encoded_test = encoded_testing_data,
                                            X_train = self.X_train,
                                            Y_train = self.Y_train,
                                            X_test = self.X_test,
                                            Y_test = self.Y_test,
                                            spaceNd=spaceNd
                                            )
        
        # ---------------------------------
        # update the dbm image
        img = Image.fromarray(np.uint8(self.dbm_plotter_gui.color_img*255))
        img.thumbnail((RESOLUTION, RESOLUTION), Image.ANTIALIAS)
        # Convert im to ImageTk.PhotoImage after window finalized
        image = ImageTk.PhotoImage(image=img)        
        self.window["-DBM IMAGE-"].update(data=image)
                    
        self.switch_visibility(["-DBM IMAGE LOADING-"], False)
        self.switch_visibility(["-DBM IMAGE-"], True)
        


# Copyright 2023 Cristian Grosu
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from sklearn.manifold import TSNE
from sklearn.decomposition import PCA
from umap import UMAP
import numpy as np

def transform_tsne(X: np.ndarray):
    """ Transforms the given data to 2D using t-SNE.

    Args:
        X (np.ndarray): The data to be transformed
    """
    X_flat = X.reshape(X.shape[0], -1)
    projection = TSNE(n_components=2, random_state=0, learning_rate="auto", init="random")
    X2d = projection.fit_transform(X_flat)
    return X2d

def transform_umap(X: np.ndarray):
    """ Transforms the given data to 2D using UMAP.

    Args:
        X (np.ndarray): The data to be transformed
    """
    X_flat = X.reshape(X.shape[0], -1)
    projection = UMAP(n_components=2, random_state=0)
    X2d = projection.fit_transform(X_flat)
    return X2d


def transform_pca(X: np.ndarray):
    """ Transforms the given data to 2D using PCA.

    Args:
        X (np.ndarray): The data to be transformed
    """
    X_flat = X.reshape(X.shape[0], -1)
    projection = PCA(n_components=2, random_state=0)
    X2d = projection.fit_transform(X_flat)
    return X2d

PROJECTION_METHODS = {
    "t-SNE": transform_tsne,
    "PCA": transform_pca,
    "UMAP": transform_umap
}



# Copyright 2022 Cristian Grosu
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from keras.datasets import mnist
import os

from .. import Logger

def import_mnist_dataset():
    """Imports the MNIST dataset from keras.datasets.mnist

    Returns:
        (X_train, Y_train), (X_test, Y_test): The train and test sets
    """
    console = Logger(name="MNIST dataset importer")
    (train_X, train_y), (test_X, test_y) = mnist.load_data()
    console.log("MNIST dataset imported")
    console.log(f"Train set: {train_X.shape}")
    console.log(f"Test set: {test_X.shape}")
    return (train_X, train_y), (test_X, test_y)

def import_csv_dataset(file_path:str, labels_index:int = 0, headers:bool = False, 
                       separator:str=",", limit:int = None, shape:tuple = (28, 28)):        
    """Imports a dataset from a csv file

    Args:
        file_path (str): The file path
        labels_index (int, optional): The index of the column with the data labels. Defaults to 0.
        headers (bool, optional): If headers are present in the file set to True. Defaults to False.
        separator (str, optional): The file data separator. Defaults to ",".
        limit (_type_, optional): The limit of data points to be loaded. Defaults to None.
        shape (tuple, optional): The data points shape. Defaults to (28, 28).

    Returns:
        X, Y (np.ndarray, np.ndarray): The data points and labels
    """
    
    if not os.path.exists(file_path):
        print("File not found")
        return None, None
    
    if not (file_path.endswith(".csv") or file_path.endswith(".txt")):
        print("File format not supported")
        return None, None
    
    try:
        with open(file_path, "r") as f:
            if headers:
                f.readline()
            lines = f.readlines()
            
            if limit is not None and limit < len(lines):
                lines = lines[:limit]
            
            lines = [line.strip().split(separator) for line in lines]
            X = [line[:labels_index] + line[labels_index+1:] for line in lines]
            Y = [line[labels_index] for line in lines]
        
        X, Y = np.array(X).astype("float32"), np.array(Y).astype("int")    
        X = X.reshape(X.shape[0], *shape)
        X /= 255
        return X, Y
    except Exception as e:
        print(e)
        return None, None    


# Copyright 2023 Cristian Grosu
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf
import numpy as np
import matplotlib.pyplot as plt

from ..DBM import DBM
from ..utils import import_mnist_dataset


def DBM_usage_example():
    
    # import the dataset
    (X_train, Y_train), (X_test, Y_test) = import_mnist_dataset()
    
    # if needed perform some preprocessing
    SAMPLES_LIMIT = 5000
    X_train = X_train.astype('float32')
    X_test = X_test.astype('float32')
    X_train /= 255
    X_test /= 255
    X_train, Y_train = X_train[:int(0.7*SAMPLES_LIMIT)], Y_train[:int(0.7*SAMPLES_LIMIT)]
    X_test, Y_test = X_test[:int(0.3*SAMPLES_LIMIT)], Y_test[:int(0.3*SAMPLES_LIMIT)]
    
    # get the number of classes
    num_classes = np.unique(Y_train).shape[0]
    
    # create a classifier
    classifier = tf.keras.models.Sequential([
            tf.keras.layers.Flatten(),
            tf.keras.layers.Dense(num_classes, activation=tf.nn.softmax)
    ], name = "classifier")
    
    # create the DBM
    dbm = DBM(classifier=classifier)
    
    # use the DBM to get the decision boundary map, if you don't have the 2D projection of the data
    # the DBM will get it for you, you just need to specify the projection method you would like to use (t-SNE, PCA or UMAP)
    img, img_confidence, _, _, _, _ = dbm.generate_boundary_map(X_train, Y_train, 
                                                                X_test, Y_test, 
                                                                resolution=256,
                                                                projection="t-SNE")
    
    # if you have the 2D projection of the data, you can use the following function to get the decision boundary map
    """
    X2d_train, X2d_test = None, None # get the 2D projection of the data by yourself
    img, img_confidence, _, _, _, _ = dbm.generate_boundary_map(X_train, Y_train, X_test, Y_test,
                                                          X2d_train=X2d_train, X2d_test=X2d_test, 
                                                          resolution=100)
    """                                                                  
    
    # make the decision boundary map pretty, by adding the colors and the confidence
    COLORS_MAPPER = {
        -2: [0,0,0], # setting original test data to black
        -1: [1,1,1], # setting original train data to white
        0: [1,0,0], 
        1: [0,1,0], 
        2: [0,0,1], 
        3: [1,1,0], 
        4: [0,1,1], 
        5: [1,0,1], 
        6: [0.5,0.5,0.5], 
        7: [0.5,0,0], 
        8: [0,0.5,0], 
        9: [0,0,0.5]
    }

    color_img = np.zeros((img.shape[0], img.shape[1], 4))
    for i, j in np.ndindex(img.shape):       
        color_img[i,j] = COLORS_MAPPER[img[i,j]] + [img_confidence[i,j]]
    
    # plot the decision boundary map
    plt.title("Decision boundary map")
    plt.axis("off")
    plt.imshow(color_img)
    plt.show()
    
    
    # use the dbm to get the inverse projection errors
    img_inverse_projection_errors = dbm.generate_inverse_projection_errors()
    # plot the inverse projection errors
    fig, ax = plt.subplots()
    ax.set_title("Inverse projection errors")
    ax.xaxis.set_visible(False)
    ax.yaxis.set_visible(False)
    img_ax = ax.imshow(img_inverse_projection_errors)
    fig.colorbar(img_ax, ax=ax)
    plt.show()
                                                           
    # use the dbm to get the projection errors
    img_projection_errors = dbm.generate_projection_errors()
    # plot the projection errors
    fig, ax = plt.subplots()
    ax.set_title("Projection errors")
    ax.xaxis.set_visible(False)
    ax.yaxis.set_visible(False)
    img_ax = ax.imshow(img_projection_errors)
    fig.colorbar(img_ax, ax=ax)
    plt.show()
    


# Copyright 2023 Cristian Grosu
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time
from .. import LoggerInterface

def track_time_wrapper(logger: LoggerInterface = None):
    """ Decorator that tracks the time it takes for a function to run.

    Args:
        logger (LoggerInterface): The logger to use. If None, prints to stdout.
    """
    def function_wrapper(func):
        def wrapper(*args, **kwargs):
            start = time.time()
            result = func(*args, **kwargs)
            end = time.time()
            if logger is None:
                print(f"{func.__name__} took {end-start} seconds")
            else:
                logger.log(f"{func.__name__} took {end-start} seconds")
            return result
        return wrapper
    return function_wrapper



# Copyright 2023 Cristian Grosu
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from math import ceil, floor, sqrt
from queue import PriorityQueue
from scipy import interpolate
import dask.array as da

from .tools import get_decode_pixel_priority, get_inv_proj_error

from ..utils import track_time_wrapper
from ..Logger import Logger, LoggerInterface

DBM_DEFAULT_RESOLUTION = 256
time_tracker_console = Logger(name="Decision Boundary Mapper - DBM", info_color="cyan", show_init=False)

class DBMInterface:
    """ Decision Boundary Mapper Interface
    
    Methods to be implemented by the class that implements the DBMInterface:
        fit (X_train, Y_train, X_test, Y_test, epochs=10, batch_size=128, load_folder = None)
        generate_boundary_map (X_train, Y_train, X_test, Y_test, train_epochs=10, train_batch_size=128, resolution=DBM_DEFAULT_RESOLUTION)
        _predict2dspace_ (X)
    
    Example of usage: 
        class SuperBoundaryMapper(DBMInterface):
            def fit (....):
                # To be implemented
            def generate_boundary_map (....):
                # To be implemented
            def _predict2dspace_ (....):
                # To be implemented
    """    
    
    def __init__(self, classifier, logger:LoggerInterface=None):
        """Initializes the classifier and the logger

        Args:
            classifier (Any): The classifier to be used
            logger (LoggerInterface, optional): The logger for the outputting info messages. Defaults to None.
        """
        if logger is None:
            self.console = Logger(name="Decision Boundary Mapper - DBM")
        else:
            self.console = logger
        
        self.console.log("Loaded classifier: " + classifier.name + "")
        self.classifier = classifier
        
    def fit(self, X_train:np.ndarray, Y_train:np.ndarray, 
            X_test:np.ndarray, Y_test:np.ndarray,
            epochs:int=10, batch_size:int=128, load_folder:bool = None):
        """ Trains the classifier on the given data set.

        Args:
            X_train (np.ndarray): Training data set
            Y_train (np.ndarray): Training data labels
            X_test (np.ndarray): Testing data set
            Y_test (np.ndarray): Testing data labels
            epochs (int, optional): The number of epochs for which the DBM is trained. Defaults to 10.
            batch_size (int, optional): Train batch size. Defaults to 128.
        """
        pass
    
    def generate_boundary_map(self, 
                              X_train:np.ndarray, Y_train:np.ndarray, 
                              X_test:np.ndarray, Y_test:np.ndarray,
                              train_epochs:int=10, train_batch_size:int=128,
                              resolution:int=DBM_DEFAULT_RESOLUTION,
                              use_fast_decoding:bool=False,
                              projection:str=None
                              ):
        """ Generates a 2D boundary map of the classifier's decision boundary.

        Args:
            X_train (np.ndarray): Training data set
            Y_train (np.ndarray): Training data labels
            X_test (np.ndarray): Testing data set
            Y_test (np.ndarray): Testing data labels
            train_epochs (int, optional): The number of epochs for which the DBM is trained. Defaults to 10.
            train_batch_size (int, optional): Train batch size. Defaults to 128.
            show_predictions (bool, optional): If set to true 10 prediction examples are shown. Defaults to True.
            resolution (int, optional): _description_. Defaults to DBM_DEFAULT_RESOLUTION = 256.
            use_fast_decoding (bool, optional): If set to true the fast decoding algorithm is used. Defaults to False.
            projection (str, optional): The projection to be used for the 2D space. Defaults to None.
        Returns:
            np.array: A 2D numpy array with the decision boundary map
        
        """
        pass
    
    def _predict2dspace_(self, X2d: np.ndarray):
        """ Predicts the labels for the given 2D data set.

        Args:
            X2d (np.ndarray): The 2D data set
        
        Returns:
            predicted_labels (np.array): The predicted labels for the given 2D data set
            predicted_confidences (np.array): The predicted probabilities for the given 2D data set
            spaceNd (np.array): The decoded nD space
        """
        pass
    
    @track_time_wrapper(logger=time_tracker_console)
    def generate_inverse_projection_errors(self, Xnd: np.ndarray = None):
        """ Calculates the inverse projection errors of the given data.

        Args:
            Xnd (np.array): The nd inverse projection of the data.
        
        Returns:
            errors (np.ndarray): The inverse projection errors matrix of the given data. (resolution x resolution)
        """
        if Xnd is None:
            if self.spaceNd is None:
                self.console.error("The nD data is not set, try to call the method 'generate_boundary_map' first.")
                raise Exception("The nD data is not set, try to call the method 'generate_boundary_map' first.")
            Xnd = self.spaceNd
            
        self.console.log("Calculating the inverse projection errors of the given data")
        errors = np.zeros(Xnd.shape[:2])
        for i in range(Xnd.shape[0]):
            for j in range(Xnd.shape[1]):
                errors[i,j] = get_inv_proj_error(i,j, Xnd)
                
        # normalizing the errors to be in the range [0,1]
        errors = (errors - np.min(errors)) / (np.max(errors) - np.min(errors))
        return errors
    
    @track_time_wrapper(logger=time_tracker_console)
    def _get_img_dbm_(self, resolution:int):
        """ This function generates the 2D image of the boundary map using the trained autoencoder and classifier.

        Args:
            resolution (int): The resolution of the 2D image to be generated 

        Returns:
            img (np.array): The 2D image of the boundary map
            img_confidence (np.array): The confidence of map of each pixel of the 2D image of the boundary map
            img_space_Nd (np.array): The nD space points
      
        Example:
            >>> img, img_confidence, space2d, spaceNd = self._get_img_dbm_(resolution = 100)
        """
        space2d = np.array([(i / resolution, j / resolution) for i in range(resolution) for j in range(resolution)])
        self.console.log("Predicting labels for the 2D boundary mapping using the nD data and the trained classifier...")
        predicted_labels, predicted_confidence, spaceNd = self._predict2dspace_(space2d)
        img = predicted_labels.reshape((resolution, resolution))
        img_confidence = predicted_confidence.reshape((resolution, resolution))
        img_space_Nd = spaceNd.reshape((resolution, resolution, -1))
        return (img, img_confidence, img_space_Nd)
    
    @track_time_wrapper(logger=time_tracker_console)
    def _get_img_dbm_fast_(self, resolution:int, computational_budget=None, interpolation_method:str="cubic"):
        """
        This function generates the 2D image of the boundary map. It uses a fast algorithm to generate the image.
        
        ATTENTION: Highly recommended to use resolution = 2 ^ n, where n is an integer number (powers of 2).
        
        Args:
            resolution (int): the resolution of the 2D image to be generated
            computational_budget (int, optional): The computational budget to be used. Defaults to None.
            interpolation_method (str, optional): The interpolation method to be used for the interpolation of sparse data generated by the fast algorithm.
                                                  Defaults to "cubic". The options are: "nearest", "linear", "cubic"
        Returns:
            img, img_confidence: The 2D image of the boundary map and a image with the confidence for each pixel
            spaceNd: The nD space points
        Example:
            >>> img, img_confidence, image_space_Nd = self._get_img_dbm_fast_(resolution=32, computational_budget=1000)
        """
        # ------------------------------------------------------------
        # Setting the initial parameters
        INITIAL_RESOLUTION = 32
        NEIGHBORS_NUMBER = 4
        
        if computational_budget is None:
            computational_budget = resolution * resolution
        
        assert(computational_budget > INITIAL_RESOLUTION * INITIAL_RESOLUTION)
        
        if (resolution % INITIAL_RESOLUTION != 0):
            self.console.warn(f"The required resolution is not a multiple of the initial resolution ({INITIAL_RESOLUTION} x {INITIAL_RESOLUTION})")
            self.console.log("The resolution will be set to the closest multiple of the initial resolution")
            resolution = int(resolution / INITIAL_RESOLUTION) * INITIAL_RESOLUTION   
        
        window_size = int(resolution / INITIAL_RESOLUTION)
        # ------------------------------------------------------------
        
        # ------------------------------------------------------------
        img = np.zeros((resolution, resolution))
        priority_queue = PriorityQueue()
        # ------------------------------------------------------------
                
        # ------------------------------------------------------------
        # generate the initial points
        space2d = [((i * window_size + window_size / 2 - 0.5) / resolution, (j * window_size + window_size / 2 - 0.5) / resolution) for i in range(INITIAL_RESOLUTION) for j in range(INITIAL_RESOLUTION)]
        
        predicted_labels, predicted_confidence, predicted_spaceNd = self._predict2dspace_(space2d)
        predicted_spaceNd = predicted_spaceNd.reshape((INITIAL_RESOLUTION, INITIAL_RESOLUTION, -1))
        # initialize the nD space
        
        computational_budget -= INITIAL_RESOLUTION * INITIAL_RESOLUTION
        
        predictions = predicted_labels.reshape((INITIAL_RESOLUTION, INITIAL_RESOLUTION))
        confidences = predicted_confidence.reshape((INITIAL_RESOLUTION, INITIAL_RESOLUTION))
        spaceNd = predicted_spaceNd.reshape((INITIAL_RESOLUTION, INITIAL_RESOLUTION, -1))
        
        img_space_Nd = np.zeros((resolution, resolution, spaceNd.shape[2]))
        
        # fill the initial points in the 2D image
        for i in range(INITIAL_RESOLUTION):
            for j in range(INITIAL_RESOLUTION):
                x0, x1, y0, y1 = i * window_size, (i+1) * window_size, j * window_size, (j+1) * window_size
                img[x0:x1+ 1, y0:y1 + 1] = predictions[i,j]
                img_space_Nd[x0:x1+ 1, y0:y1 + 1] = spaceNd[i,j]
        # -------------------------------------
        
        # collecting the indexes of the actual computed pixels in the 2D image and the confidence of each pixel
        # creating an artificial border for the 2D confidence image
        confidence_map = []
        
        if interpolation_method != "nearest":
            border_indices = [(i * window_size + window_size / 2 - 0.5, 0) for i in range(INITIAL_RESOLUTION)] + \
                            [(i * window_size + window_size / 2 - 0.5, resolution - 1) for i in range(INITIAL_RESOLUTION)] + \
                            [(0, i * window_size + window_size / 2 - 0.5) for i in range(-1, INITIAL_RESOLUTION+1)] + \
                            [(resolution - 1, i * window_size + window_size / 2 - 0.5) for i in range(-1, INITIAL_RESOLUTION + 1)]
            
            space2d_border = [(i / resolution, j / resolution) for (i,j) in border_indices]
            _, confidences_border, _ = self._predict2dspace_(space2d_border)
            computational_budget -= len(space2d_border)
            confidence_map = [(i, j, conf) for (i,j),conf in zip(border_indices, confidences_border)]
                    
        # analyze the initial points and generate the priority queue        
        for i in range(INITIAL_RESOLUTION):
            for j in range(INITIAL_RESOLUTION):
                x, y = i * window_size + window_size / 2 - 0.5, j * window_size + window_size / 2 - 0.5
                confidence_map.append((x,y,confidences[i,j]))                
                priority = get_decode_pixel_priority(img, x, y, window_size, predictions[i,j])
                if priority != -1:
                    priority_queue.put((priority, (window_size, x, y)))

        # -------------------------------------
        # start the iterative process of filling the image
        while computational_budget > 0 and not priority_queue.empty():    
            # take the highest priority task
            priority, item = priority_queue.get()
            
            # getting all the items with the same priority
            items = [item]
            if not priority_queue.empty():
                next_priority, next_item = priority_queue.get()
                while priority == next_priority:
                    items.append(next_item)
                    if priority_queue.empty():
                        break
                    next_priority, next_item = priority_queue.get()
                if priority != next_priority:
                    priority_queue.put((next_priority, next_item))
            
            
            space2d, indices = [], []
            single_points_space, single_points_indices = [], []
            valid_items = []
            
            for item in items:            
                (window_size, i, j) = item
                
                if window_size == 1:
                    single_points_indices += [(int(i), int(j))]
                    single_points_space += [(i / resolution, j / resolution)]
                    continue
                
                window_size = window_size / NEIGHBORS_NUMBER
                neighbors = [(i - window_size, j - window_size), (i - window_size, j + window_size), (i + window_size, j - window_size), (i + window_size, j + window_size)]
                space2d += [(x / resolution, y / resolution) for (x, y) in neighbors]
                indices += neighbors                                                
                valid_items.append(item)
            
            
            space = space2d + single_points_space
            
            # check if the computational budget is enough and update it
            if computational_budget - len(space) < 0:
                self.console.warn("Computational budget exceeded, stopping the process")
                break
            
            computational_budget -= len(space)
            
            # decode the space
            predicted_labels, predicted_confidence, predicted_spaceNd = self._predict2dspace_(space)
            predicted_spaceNd = predicted_spaceNd.reshape((len(space), -1))
            # copy the image to a new one, the new image will be updated with the new labels, the old one will be used to calculate the priorities
            new_img = np.copy(img)            
            
            # fill the new image with the single points
            single_points_labels = predicted_labels[len(space2d):]
            single_points_confidences  = predicted_confidence[len(space2d):]
            single_points_spaceNd = predicted_spaceNd[len(space2d):]
            
            for i in range(len(single_points_indices)):
                new_img[single_points_indices[i]] = single_points_labels[i]
                img_space_Nd[single_points_indices[i]] = single_points_spaceNd[i]
                confidence_map.append((single_points_indices[i][0], single_points_indices[i][1], single_points_confidences[i]))                
                           
            predicted_labels = predicted_labels[:len(space2d)]
            predicted_confidence = predicted_confidence[:len(space2d)]
            predicted_spaceNd = predicted_spaceNd[:len(space2d)] 
            
            # fill the new image with the new labels and update the priority queue
            for index in range(len(valid_items)):
                (window_size, i, j) = valid_items[index]
                neighbors = indices[index*NEIGHBORS_NUMBER:(index+1)*NEIGHBORS_NUMBER]
                neighbors_labels = predicted_labels[index*NEIGHBORS_NUMBER:(index+1)*NEIGHBORS_NUMBER]
                confidences = predicted_confidence[index*NEIGHBORS_NUMBER:(index+1)*NEIGHBORS_NUMBER]
                spaceNd = predicted_spaceNd[index*NEIGHBORS_NUMBER:(index+1)*NEIGHBORS_NUMBER]
                
                new_window_size = window_size / NEIGHBORS_NUMBER
                
                for (x, y), label, conf, pointNd in zip(neighbors, neighbors_labels, confidences, spaceNd):
                    # all the pixels in the window are set to the same label
                    # starting from the top left corner of the window
                    # ending at the bottom right corner of the window
                    # the +1 is because the range function is not inclusive
                    confidence_map.append((x, y, conf))
                    if new_window_size >= 1:
                        x0, x1, y0, y1 = ceil(x-new_window_size), floor(x + new_window_size), ceil(y - new_window_size), floor(y + new_window_size)
                        new_img[x0:x1 + 1, y0:y1 + 1] = label
                        img_space_Nd[x0:x1 + 1, y0:y1 + 1] = pointNd
                    else:
                        new_img[int(x), int(y)] = label
                        img_space_Nd[int(x), int(y)] = pointNd
                
                # update the priority queue after the window has been filled
                for (x, y), label in zip(neighbors, neighbors_labels):
                    priority = get_decode_pixel_priority(img, x, y, 2 * new_window_size, label)
                    if priority != -1:
                        priority_queue.put((priority, (2 * new_window_size, x, y)))
            
            # update the image        
            img = new_img
        
        # generating the confidence image using interpolation based on the confidence map
        img_confidence = self._generate_interpolated_image_(sparse_map=confidence_map,
                                                            resolution=resolution,
                                                            method=interpolation_method).T
        
        return img, img_confidence, img_space_Nd

    def _generate_interpolated_image_(self, sparse_map, resolution, method='cubic'):
        """A private method that uses interpolation to generate the values for the 2D space image
           The sparse map is a list of tuples (x, y, data)
           The sparse map represents a structured but non uniform grid of data values
           Therefore usual rectangular interpolation methods are not suitable
           For the interpolation we use the scipy.interpolate.griddata function with the cubic method
        Args:
            sparse_map (list): a list of tuples (x, y, data) where x and y are the coordinates of the pixel and data is the data value
            resolution (int): the resolution of the image we want to generate (the image will be a square image)
            method (str, optional): The method to be used for the interpolation. Defaults to 'cubic'. Available methods are: 'nearest', 'linear', 'cubic'

        Returns:
            np.array: an array of shape (resolution, resolution) containing the data values for the 2D space image
        """
        X, Y, Z = [], [], []
        for (x, y, z) in sparse_map:
            X.append(x)
            Y.append(y)
            Z.append(z)
        X, Y, Z = np.array(X), np.array(Y), np.array(Z)
        xi = np.linspace(0, resolution-1, resolution)
        yi = np.linspace(0, resolution-1, resolution)
        return interpolate.griddata((X, Y), Z, (xi[None,:], yi[:,None]), method=method)

    @track_time_wrapper(logger=time_tracker_console)
    def _generate_interpolation_rbf_(self, sparse_map, resolution, function='linear'):
        """A private method that uses interpolation to generate the values for the 2D space image
           The sparse map is a list of tuples (x, y, data) where x, y and data are in the range [0, 1]
        
        Args:
            sparse_map (np.ndarray): a list of tuples (x, y, data) where x and y are the coordinates of the pixel and data is the data value
            resolution (int): the resolution of the image we want to generate (the image will be a square image)
            function (str, optional): Defaults to 'euclidean'.
        """
        X, Y, Z = [], [], []
        for (x, y, z) in sparse_map:
            X.append(x)
            Y.append(y)
            Z.append(z)

        rbf = interpolate.Rbf(X, Y, Z, function=function) 
        ti = np.linspace(0, 1, resolution)
        xx, yy = np.meshgrid(ti, ti)
        # using dask to parallelize the computation of the rbf function
        # the rbf function is applied to each pixel of the image
        # the image is divided into chunks and each chunk is processed in parallel
        # the result is then merged together
        # the number of chunks is equal to the number of cores
        cores = 4
        ix = da.from_array(xx, chunks=(1, cores))
        iy = da.from_array(yy, chunks=(1, cores))
        iz = da.map_blocks(rbf, ix, iy)
        zz = iz.compute()
        return zz


# Copyright 2023 Cristian Grosu
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2023 Cristian Grosu
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.offsetbox import OffsetImage, AnnotationBbox, TextArea
from PIL import Image
import numpy as np
import PySimpleGUI as sg

from .. import Logger

def draw_figure_to_canvas(canvas, figure):
    figure_canvas_agg = FigureCanvasTkAgg(figure, canvas)
    figure_canvas_agg.draw()
    figure_canvas_agg.get_tk_widget().pack(side='top', fill='both', expand=1)
    return figure_canvas_agg


def generate_color_mapper():
    colors_mapper = {
        -2: [0,0,0], # setting original test data to black
        -1: [1,1,1], # setting original train data to white
        0: [1,0,0], 
        1: [0,1,0], 
        2: [0,0,1], 
        3: [1,1,0], 
        4: [0,1,1], 
        5: [1,0,1], 
        6: [0.5,0.5,0.5], 
        7: [0.5,0,0], 
        8: [0,0.5,0], 
        9: [0,0,0.5]
    }
    # setting the rest of the colors 
    for i in range(10, 100):
        colors_mapper[i] = [i/200,i/200,i/200]
    return colors_mapper

# Generating initial settings
COLORS_MAPPER = generate_color_mapper()
APP_FONT = 'Helvetica 12'
TITLE = "Decision Boundary Map & Errors"
WINDOW_SIZE = (1550, 950)
BLACK_COLOR = "#252526"
BUTTON_PRIMARY_COLOR = "#007acc"
WHITE_COLOR = "#ffffff"
RIGHTS_MESSAGE = "© 2023 Cristian Grosu. All rights reserved."
RIGHTS_MESSAGE_2 = "Made by Cristian Grosu for Utrecht University Master Thesis in 2023"


class DBMPlotterGUI:
    def __init__ (self, 
                  dbm_model,
                  img, 
                  img_confidence,
                  X_train, 
                  Y_train, 
                  X_test, 
                  Y_test,
                  encoded_train, 
                  encoded_test,                                      
                  spaceNd, logger=None):
        
        
        if logger is None:
            self.console = Logger(name="DBMPlotterGUI")
        else:
            self.console = logger
        
        self.dbm_model = dbm_model 
        self.img = img
        self.img_confidence = img_confidence
        self.X_train = X_train
        self.Y_train = Y_train
        self.X_test = X_test
        self.Y_test = Y_test
        self.encoded_train = encoded_train
        self.encoded_test = encoded_test                                     
        self.spaceNd = spaceNd.reshape(spaceNd.shape[0], spaceNd.shape[1], X_train.shape[1], X_train.shape[2])
        self.color_img, self.legend = self._build_2D_image_(img, img_confidence)
        self.train_mapper, self.test_mapper = self._generate_encoded_mapping_()
        self.inverse_projection_errors = self.dbm_model.generate_inverse_projection_errors()
        
        # --------------------- Plotter related ---------------------                
        self.fig, self.dbm_ax, self.proj_errs_ax, self.inv_proj_errs_ax = self._build_plot_()        
        self._build_annotation_mapper_()
        
        ax_img = self.inv_proj_errs_ax.imshow(self.inverse_projection_errors)        
        self.fig.colorbar(ax_img, ax=self.inv_proj_errs_ax)
        self.fig.legend(handles=self.legend, borderaxespad=0. )
        # --------------------- Plotter related ---------------------
        self.proj_errs_computed = False    
        
        
    def _initialize_gui_(self):        
        # --------------------- GUI related ---------------------
        self.window = self._build_GUI_()
        self.canvas, self.fig_agg = self._build_canvas_(self.fig, key = "-DBM CANVAS-")
        
        self.draw_dbm_img()    
        # --------------------- GUI related ---------------------        
            
    def _get_GUI_layout_(self):   
        buttons = []
           
        if not self.proj_errs_computed: 
            buttons.append(sg.Button('Compute Projection Errors', font=APP_FONT, expand_x=True, key="-COMPUTE PROJECTION ERRORS-", button_color=(WHITE_COLOR, BUTTON_PRIMARY_COLOR)))             
      
        layout = [                                  
                    [
                        [sg.Canvas(key='-DBM CANVAS-', expand_x=True, expand_y=True, pad=(0,0))],                        
                    ], 
                    buttons,    
                    [sg.Text(RIGHTS_MESSAGE)],
                    [sg.Text(RIGHTS_MESSAGE_2)]              
                ]
        return layout
        
    def _build_GUI_(self):        
        window = sg.Window(title = TITLE, 
                           layout = self._get_GUI_layout_(), 
                           size=WINDOW_SIZE, 
                           resizable=True,
                           element_justification='center',
                           )
        window.finalize()
        return window
    
    def start(self):
        self._initialize_gui_()
        while True:
            event, values = self.window.read()
            
            if event == "Exit" or event == sg.WIN_CLOSED:
                break
        
            self.handle_event(event, values)
        
        self.stop()
    
    def stop(self):
        self.console.log("Closing the application...")
        self.window.close()
    
    def handle_event(self, event, values):
        EVENTS = {
            "-COMPUTE PROJECTION ERRORS-": self.handle_compute_projection_errors_event,
        }
        
        EVENTS[event](event, values)    
    
    def _build_2D_image_(self, img, img_confidence, class_name_mapper = lambda x: str(x), colors_mapper = COLORS_MAPPER):
        """Combines the img and the img_confidence into a single image.

        Args:
            img (np.ndarray): Label image.
            img_confidence (np.ndarray): Confidence image.
            class_name_mapper (function, optional): Describes how to map the data labels. Defaults to lambdax:str(x).
            colors_mapper (dict, optional): Mapper of labels to colors. Defaults to COLORS_MAPPER.

        Returns:
            np.ndarray: The combined image.
            patches: The legend patches.
        """
        color_img = np.zeros((img.shape[0], img.shape[1], 4))
                
        for i, j in np.ndindex(img.shape):
            color_img[i,j] = colors_mapper[img[i,j]] + [img_confidence[i,j]]
        
        values = np.unique(img)
        
        patches = []
        for value in values:
            color = colors_mapper[value]
            if value==-1:
                label = "Original train data"
            elif value==-2:
                label = "Original test data"
            else:
                label = f"Value region: {class_name_mapper(value)}"

            patches.append(mpatches.Patch(color=color, label=label))
        
        return color_img, patches
   
    def _generate_encoded_mapping_(self):
        """Generates a mapping of the encoded data to the original data.

        Returns:
            train_mapper (dict): Mapping of the encoded train data to the original train data.
            test_mapper (dict): Mapping of the encoded test data to the original test data.
        """
        train_mapper = {}
        for k in range(len(self.encoded_train)):
            [x, y] = self.encoded_train[k]
            train_mapper[f"{x} {y}"] = k
    
        test_mapper = {}
        for k in range(len(self.encoded_test)):
            [x, y] = self.encoded_test[k]
            test_mapper[f"{x} {y}"] = k
        
        return train_mapper, test_mapper

    def _build_plot_(self):                   
        fig = plt.figure(figsize = (1, 1))
        ax = fig.add_subplot(121)
        ax1 = fig.add_subplot(222)
        ax2 = fig.add_subplot(224)
        ax.set_axis_off()
        ax1.set_axis_off()
        ax2.set_axis_off()       
        return fig, ax, ax1, ax2
    
    def _build_canvas_(self, fig, key):        
        canvas = self.window[key].TKCanvas
        fig_agg = draw_figure_to_canvas(canvas, fig)
        return canvas, fig_agg
    
    def _build_annotation_mapper_(self):
        """ Builds the annotation mapper.
            This is used to display the data point label when hovering over the decision boundary.
        """
        image = OffsetImage(self.X_train[0], zoom=2, cmap="gray")
        label = TextArea("Data point label: None")

        xybox=(50., 50.)
        annImage = AnnotationBbox(image, (0,0), xybox=xybox, xycoords='data',boxcoords="offset points",  pad=0.1,  arrowprops=dict(arrowstyle="->"))
        annImage_ax1 = AnnotationBbox(image, (0,0), xybox=xybox, xycoords='data', boxcoords="offset points",  pad=0.1,  arrowprops=dict(arrowstyle="->"))
        annImage_ax2 = AnnotationBbox(image, (0,0), xybox=xybox, xycoords='data', boxcoords="offset points",  pad=0.1,  arrowprops=dict(arrowstyle="->"))

        annLabels = AnnotationBbox(label, (0,0), xybox=xybox, xycoords='data', boxcoords="offset points",  pad=0.3,  arrowprops=dict(arrowstyle="->"))

        self.dbm_ax.add_artist(annImage)
        self.dbm_ax.add_artist(annLabels)
        self.inv_proj_errs_ax.add_artist(annImage_ax1)
        self.proj_errs_ax.add_artist(annImage_ax2)
        annImage.set_visible(False)
        annImage_ax1.set_visible(False)
        annImage_ax2.set_visible(False)
        annLabels.set_visible(False)

        fig_width, fig_height = self.fig.get_size_inches() * self.fig.dpi
            
        def display_annotation(event):
            """Displays the annotation box when hovering over the decision boundary based on the mouse position."""
            if event.inaxes == None:
                return
        
            j, i = int(event.xdata), int(event.ydata)
            
            # change the annotation box position relative to mouse.
            ws = (event.x > fig_width/2.)*-1 + (event.x <= fig_width/2.) 
            hs = (event.y > fig_height/2.)*-1 + (event.y <= fig_height/2.)
            annImage.xybox = (50. * ws, 50. * hs)
            annLabels.xybox = (-50. * ws, 50. *hs)
            # make annotation box visible
            annImage.set_visible(True)
            annImage_ax1.set_visible(True)
            annImage_ax2.set_visible(True)
            # place it at the position of the event scatter point
            annImage.xy = (j, i)
            annImage_ax1.xy = (j, i)
            annImage_ax2.xy = (j, i)
            
            x_data, y_data = find_data_point(i,j)                            
            image.set_data(x_data)
            
            if y_data is not None:    
                annLabels.set_visible(True)
                annLabels.xy = (j, i)
                label.set_text(f"Label: {y_data}")
            else:
                annLabels.set_visible(False)

            self.fig.canvas.draw_idle()
                
        def onclick(event):
            """ Open the data point in a new window when clicked on a pixel in training or testing set based on mouse position."""
            if event.inaxes == None:
                return
            
            self.console.log("Clicked on: " + str(event.xdata) + ", " + str(event.ydata))
            j, i = int(event.xdata), int(event.ydata)

            if self.img[i,j] >= 0 or self.img[i,j] < -2:
                self.console.log("Data point not in training or testing set")
                return
            x, y = find_data_point(i, j)
            self.plot_data_point(x, y)
            
            
        def find_data_point(i, j):
            if self.img[i,j] == -1:
                k = self.train_mapper[f"{i} {j}"]
                #self.console.log("Found point in train set")
                return self.X_train[k], self.Y_train[k]
            
            if self.img[i,j] == -2:
                k = self.test_mapper[f"{i} {j}"]
                #self.console.log("Found point in test set")
                return self.X_test[k], self.Y_test[k]        
            
            point = self.spaceNd[i,j]
            #self.console.log("Found point in value region")
            return point, None
            
            
        self.fig.canvas.mpl_connect('motion_notify_event', display_annotation)           
        self.fig.canvas.mpl_connect('button_press_event', onclick)
       
    def plot_data_point(self, data, label):
        """Plots the data point in a new window.

        Args:
            data (np.ndarray): the data point to be displayed
            label (str): the label of the data point
        """
        img = Image.fromarray(np.uint8(data * 255))
        img.thumbnail((100, 100), Image.ANTIALIAS)
        img.show(title=f"Data point label: {label}")
    
    def draw_dbm_img(self):
        # clear the figure
        #self.dbm_fig.clf()   
        self.fig_agg.get_tk_widget().forget()
             
        # update the figure
        self.dbm_ax.set_title("Decision Boundary Mapper")
        self.proj_errs_ax.set_title("Projection Errors")
        self.inv_proj_errs_ax.set_title("Inverse Projection Errors")
        
        self.dbm_ax.imshow(self.color_img)
        self.inv_proj_errs_ax.imshow(self.inverse_projection_errors)
        
        if not self.proj_errs_computed:
            self.proj_errs_warning_text = self.proj_errs_ax.text(0.5, 0.5, "To compute projection errors you should click the button.\nThis might take about 30 seconds", transform=self.proj_errs_ax.transAxes, ha="center")
        
        
        # draw the figure to the canvas
        self.fig_agg = draw_figure_to_canvas(self.canvas, self.fig)    
    
    def _draw_projection_errors_img_(self, projection_errors):
        # clear the figure
        self.fig_agg.get_tk_widget().forget()
        
        # update the figure        
        self.proj_errs_warning_text.set_visible(False)
        ax_img = self.proj_errs_ax.imshow(projection_errors)
                
        self.fig.colorbar(ax_img, ax=self.proj_errs_ax)        
        self.fig.set_size_inches(1, 1)        
        
        # draw the figure to the canvas
        self.fig_agg = draw_figure_to_canvas(self.canvas, self.fig) 

    def _set_loading_proj_errs_state_(self):
        self.window['-COMPUTE PROJECTION ERRORS-'].update(visible=False, disabled=True)        
        self.proj_errs_warning_text.set_text("Computing projection errors...\nPlease wait...")
        self.fig_agg.get_tk_widget().forget()
        self.fig_agg = draw_figure_to_canvas(self.canvas, self.fig) 

    def handle_compute_projection_errors_event(self, event, values):
        self._set_loading_proj_errs_state_()
        projection_errors = self.dbm_model.generate_projection_errors()
        self.proj_errs_computed = True
        self._draw_projection_errors_img_(projection_errors)
